
/*
 * ngx_http_oceanbase_module.h
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 * Copyright (C) Yichun Zhang (agentzh)
 *
 */
#ifndef NGX_HTTP_OCEANBASE_MODULE_H
#define NGX_HTTP_OCEANBASE_MODULE_H

#include <ngx_config.h>
#include <nginx.h>
#include <ngx_http.h>

#ifndef NGX_HTTP_GONE
#define NGX_HTTP_GONE  410
#endif

extern ngx_module_t ngx_http_oceanbase_module;

typedef struct {
    ngx_uint_t                          key;
    ngx_str_t                           sv;
    ngx_http_complex_value_t           *cv;
} ngx_oceanbase_mixed_t;

typedef struct {
    u_char                             *name;
    uint32_t                            key;
} ngx_oceanbase_http_method_t;

typedef struct {
	ngx_uint_t  gid;
	ngx_uint_t 	nid;
	ngx_uint_t	size;
} ngx_oceanbase_delivery_queue_t;

typedef struct {
	ngx_uint_t start;
	ngx_uint_t rup;
	ngx_uint_t rdown;
	ngx_uint_t mi;
	ngx_uint_t ri;
	ngx_uint_t wts;
} ngx_oceanbase_delivery_runtime_t;

typedef struct {
	char dsn[128];
	char user[128];
	char passwd[128];
} ngx_oceanbase_delivery_mdb_t;

typedef struct {
    ngx_http_upstream_conf_t             upstream;
    /* oceanbase database name */
    ngx_http_complex_value_t            *dsn;
    ngx_msec_t                           get_result_timeout;
    ngx_http_complex_value_t            *complex_target;
    ngx_uint_t							tpcc_terminal_num;
	ngx_uint_t                          tpcc_max_txn_cnt;
	ngx_oceanbase_delivery_queue_t		tpcc_delivery_queue;
	ngx_oceanbase_delivery_runtime_t    tpcc_delivery_runtime;
	ngx_oceanbase_delivery_mdb_t 		tpcc_delivery_mdb;
} ngx_http_oceanbase_loc_conf_t;


/* states for the oceanbase client state machine */
typedef enum {
    state_db_connect,
    state_db_send_query,
	state_db_get_result,
    state_db_idle
} ngx_http_oceanbase_state_t;


#endif /* NGX_HTTP_OCEANBASE_MODULE_H */
