/*
 * ngx_http_oceanbase_upstream.h
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */ 

#ifndef NGX_HTTP_OCEANBASE_UPSTREAM_H
#define NGX_HTTP_OCEANBASE_UPSTREAM_H

#include <ngx_core.h>
#include <ngx_http.h>
#include <nginx.h>
#include "ngx_http_oceanbase_module.h"
#include "ngx_tpcc_txn.h"

typedef enum {
    oceanbase_keepalive_overflow_ignore = 0,
    oceanbase_keepalive_overflow_reject
} ngx_http_oceanbase_keepalive_overflow_t;

typedef struct {
    ngx_str_t                        dsn;
    ngx_str_t                        user;
    ngx_str_t                        password;
} ngx_http_upstream_oceanbase_server_t;

typedef ngx_http_upstream_oceanbase_server_t ngx_http_upstream_oceanbase_peer_t;

typedef struct {
    ngx_http_upstream_oceanbase_peer_t   *peer;
    ngx_array_t                         *servers;
    ngx_pool_t                          *pool;
    ngx_queue_t                          free;
    ngx_queue_t                          cache;
    ngx_uint_t                           active_conns;
    ngx_uint_t                           max_cached;
    ngx_http_oceanbase_keepalive_overflow_t    overflow;
} ngx_http_upstream_oceanbase_srv_conf_t;


typedef struct {
    ngx_http_oceanbase_loc_conf_t            *loc_conf;
    ngx_http_upstream_oceanbase_srv_conf_t   *srv_conf;
    ngx_http_upstream_t                    *upstream;
    ngx_http_request_t                     *request;
    ngx_str_t                               dsn;
    ngx_http_oceanbase_state_t                state;
    oceanbase_connection                         *oceanbase_con;
    ngx_chain_t                           **last_out;
    ngx_buf_t                              *out_buf;
    ngx_buf_t                               cached;
    ngx_buf_t                               postponed;
    size_t                                  avail_out;
    ngx_uint_t                              used;
    unsigned                                failed:1;
    unsigned                                seen_stream_end:1;
} ngx_http_upstream_oceanbase_peer_data_t;


char *ngx_http_upstream_oceanbase_server(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
void *ngx_http_upstream_oceanbase_create_srv_conf(ngx_conf_t *cf);
void ngx_http_upstream_oceanbase_free_connection(ngx_log_t *log,
        ngx_connection_t *c, oceanbase_connection *obc,
        ngx_http_upstream_oceanbase_srv_conf_t *obscf);
ngx_http_upstream_srv_conf_t *ngx_http_upstream_oceanbase_add(ngx_http_request_t *r, ngx_url_t *url);

#endif /* NGX_HTTP_OCEANBASE_UPSTREAM_H */
