/*
 * ngx_tpcc_common.h is the common declaration
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */
#ifndef __NGX_TPCC_COMMON_H__
#define __NGX_TPCC_COMMON_H__
#include <math.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>

#define TV_2_MSEC(tv) (tv.tv_sec * 1000LL + tv.tv_usec / 1000)
#define TV_2_SEC(tv) (tv.tv_sec + tv.tv_usec/ 1000000)
#define TV_2_MT(tv) (tv.tv_sec * 10000LL + tv.tv_usec / 100)
#define MT_2_MSEC(mt) ((mt) * 1.0 / 10)
#define MT_2_SEC(mt) ((mt) * 1.0 / 10000)
#define MSEC_2_SEC(ms) ((ms) * 1.0 / 1000)
#define MT_2_MSEC_CEIL(mt) ((int)ceil((mt) * 1.0 / 10))

#define REPORT_TXN_INTERVAL 300

#define DEF_MSG_LEN 1024
#define DEF_SQL_LEN 1024

#define TXN_TYPE_NUM 5
#define TXN_TYPE_NUM_PLUS 6

typedef enum {NEW_ORDER, PAYMENT, DELIVERY, ORDER_STATUS, STOCK_LEVEL, DELIVERY_DEF, TXN_MENU, LOGON, TXN_UNKNOWN} TXN_TYPE;

#define NEW_ORDER_PROC "CALL NEWORDER(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
#define PAYMENT_PROC "CALL PAYMENT(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
#define ORDER_STATUS_PROC "CALL ORDERSTATUS(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
#define STOCK_LEVEL_PROC "CALL STOCKLEVEL(?,?,?,?);"

#define bool int
#define true 1
#define false 0

typedef struct _TERMINAL_STATE{
	int terminal_id;
	bool active;
}TERMINAL_STATE;

typedef struct _TERMINAL_STATES{
	int	max_terminal_cnt;
	TERMINAL_STATE terminal_state[1];
}TERMINAL_STATES;

typedef struct txn_stat{
    long long int txn_cnt;
    long long int total_in_time;
    long long int total_db_time;
    int max_in_time;
    int max_db_time;
}txn_stat;

#define MAX_TERMINAL_NUM	4000000
#define MAX_DEL_TASK_NUM	100000

#define DELIVERY_TASK_SHM	0x70000000
#define DELIVERY_TASK_SEMA  0x60000000

#define NEW_ORDER_INPUT_URI  "/tpcc/new-order-input"
#define PAYMENT_INPUT_URI "/tpcc/payment-input"
#define DELIVERY_INPUT_URI "/tpcc/delivery-input"
#define ORDER_STATUS_INPUT_URI "/tpcc/order-status-input"
#define STOCK_LEVEL_INPUT_URI "/tpcc/stock-level-input"

#define NEW_ORDER_URI  "/tpcc/new-order"
#define PAYMENT_URI "/tpcc/payment"
#define DELIVERY_URI "/tpcc/delivery"
#define ORDER_STATUS_URI "/tpcc/order-status"
#define STOCK_LEVEL_URI "/tpcc/stock-level"
#define TXN_MENU_URI "/tpcc/txn-menu"
#define LOGON_URI "/tpcc/logon"

#define DELIVERY_THREAD_POOL "delivery_thread_pool"

extern bool SQL_SUCC(int retcode);

typedef struct conn
{
    char dsn[128];
    char user[128];
    char passwd[128];
}ConnInfo;

typedef struct _delivery_task_info{
    int w_id;
    int o_carrier_id;
	long long int dtt;
    long long int enqueue_time;
	long long int pos; 
}DeliveryTask;


void get_sql_diag(SQLSMALLINT type, SQLHANDLE handle, char *state, bool *reconn);

bool get_timestamp(char * str);

bool SQL_SUCC(int ret);

bool SQL_FAIL(int ret);

bool rollback_on_error(SQLHANDLE dbc, char* msg, bool* need_reconn);
#endif
