/*
 * ngx_tpcc_delivery_thread.c is backend delivery processing routine
 * 
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */
#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/time.h>
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>
#include <unistd.h>
#include <ngx_core.h>
#include "ngx_tpcc_common.h"
#include "ngx_tpcc_delivery_thread.h"

char *get_cur_time()
{
    struct tm *lt, lt_;
    struct timeval tv;
    static __thread char cur_time_str[50] = {'\0'};

    if(gettimeofday(&tv, NULL) == 0){
        tzset();
        lt = localtime_r(&tv.tv_sec, &lt_);
        if(lt != NULL){
            sprintf(cur_time_str, "[%04d/%02d/%02d %02d:%02d:%02d.%06ld] ",
                   lt->tm_year + 1900,
                   lt->tm_mon + 1,
                   lt->tm_mday,
                   lt->tm_hour,
                   lt->tm_min,
                   lt->tm_sec,
                   tv.tv_usec);
        }else{
            sprintf(cur_time_str, "localtime_r failed errno: %d\n", errno);
        }
    }else{
       sprintf(cur_time_str, "gettimeofday failed errno: %d\n", errno);
    }

    return cur_time_str;
}

bool get_start_time_str(long start, char* start_str)
{
    struct tm *lt, lt_;
    bool ret = true;

	tzset();
    lt = localtime_r((time_t *)&start, &lt_);
    if(lt != NULL){
        sprintf(start_str, "%04d%02d%02d%02d%02d%02d",
                   lt->tm_year + 1900,
                   lt->tm_mon + 1,
                   lt->tm_mday,
                   lt->tm_hour,
                   lt->tm_min,
                   lt->tm_sec);
    }else{
        ret = false;
    }

    return ret;
}

pthread_mutex_t  g_del_mutex;
FILE * g_del_res_files[DEL_FILES_NUM] = {NULL, NULL};
DeliveryStatBuf * g_del_stat_buf = NULL;
int * g_del_rt = NULL;
RealtimeStatBuf * g_del_rts_buf = NULL;
int	g_total_conn_num_del = 0;
int g_rpt_total_conn_num_del = 0;
bool g_rpt_conn_num_first_time = true;

DeliveryServer g_del_server;
ConnInfo g_del_mdb_conn;
ConnInfo g_del_conn;
RuntimeConf g_del_runtime;
char g_del_start_str[30] = {'\0'};
int g_del_last_rpt_idx = -1;

#define DELIVERY_PROC "CALL DELIVERY(?,?,?,?,?,?,?,?,?);"

static __thread	bool g_del_thread_inited = false;	
static __thread	SQLHENV g_del_env = NULL;	
static __thread	SQLHDBC g_del_dbc = NULL;
static __thread	SQLHSTMT g_del_stmt = NULL;

static __thread NUMBER_STRUCT del_d_ids_num[DISTS_PER_WARE];
static __thread ARRAY_STRUCT del_d_ids_arr;
static __thread NUMBER_STRUCT del_o_ids_num[DISTS_PER_WARE];
static __thread ARRAY_STRUCT del_o_ids_arr;
static __thread NUMBER_STRUCT del_sums_num[DISTS_PER_WARE];
static __thread ARRAY_STRUCT del_sums_arr;
static __thread NUMBER_STRUCT del_c_ids_num[DISTS_PER_WARE];
static __thread ARRAY_STRUCT del_c_ids_arr;

static __thread int del_w_id;
static __thread int del_o_carrier_id;

static __thread int del_d_ids[DISTS_PER_WARE] = {0};
static __thread int del_o_ids[DISTS_PER_WARE] = {0};
static __thread int del_o_cnt;
static __thread int del_sums[DISTS_PER_WARE] = {0};
static __thread char del_d_date[30]= {'\0'};
static __thread int del_c_ids[DISTS_PER_WARE] = {0};
static __thread	int del_retry = 0;

SQLHANDLE g_mdb_env = NULL;
SQLHANDLE g_mdb_dbc = NULL;
SQLHANDLE g_mdb_stmt= NULL;

bool delivery_dumper(bool all);

bool init_del_statbuf()
{
	bool ret = true;
	int queue_size = MAX_DEL_STAT_BUF_SIZE;

	if(queue_size <= 0){
		ret = false;
	}else if(!(g_del_stat_buf = (DeliveryStatBuf *)malloc(sizeof(DeliveryStatBuf)
                                        + (queue_size - 1)*sizeof(DeliveryStat)))
            || !(g_del_rt = (int *)calloc(queue_size, sizeof(int)))){
		ret = false;
	}else{
		memset(g_del_stat_buf, 0, sizeof(DeliveryStatBuf) + (queue_size - 1)*sizeof(DeliveryStat));	
		g_del_stat_buf->total_num = queue_size;	
		g_del_stat_buf->min_rt = DEF_MIN_RT;
	}

	return ret;	
}

bool init_del_rts_points(int rup, int rdown, int mi, int ri)
{
	bool ret = true;
	int point_num = (rup + rdown + mi) / ri + PNT_NUM_PLUS; 
	
	if(point_num <= 0){
		ret = false;
	}else if(!(g_del_rts_buf = (RealtimeStatBuf *)malloc(sizeof(RealtimeStatBuf)
                                        + (point_num - 1)*sizeof(RealtimeStat)))){
		ret = false;
	}else{
		memset(g_del_rts_buf, 0, sizeof(RealtimeStatBuf) + (point_num - 1)*sizeof(RealtimeStat));	
		g_del_rts_buf->point_num = point_num;	
	}

	return ret;	
}

void get_delivery_summary(char *del_msg)
{
	int len = 0;

	len += sprintf(del_msg + len, "Total delivery txn(cmt/rbk/lack ord): %lld/%lld/%lld\n\n", 
										g_del_stat_buf->cmt_num + g_del_stat_buf->lack_ord_num,
										g_del_stat_buf->err_num,
										g_del_stat_buf->lack_ord_num);
	len += sprintf(del_msg + len, "Response time(max/min/avg)(s): %.3f/%.3f/%.3f\n\n",
										MSEC_2_SEC(g_del_stat_buf->max_rt),
										MSEC_2_SEC(g_del_stat_buf->min_rt),
			MSEC_2_SEC(g_del_stat_buf->total_rt) / (g_del_stat_buf->cmt_num + g_del_stat_buf->lack_ord_num));
	len += sprintf(del_msg + len, "Average queue time(s): %.3f\n\n",
			MSEC_2_SEC(g_del_stat_buf->total_qt) / (g_del_stat_buf->cmt_num + g_del_stat_buf->lack_ord_num));
}


// record transaction stats to metadb
bool init_metadb()
{	 
	char out_state[DEF_MSG_LEN] = {'\0'};
	SQLRETURN retcode = SQL_SUCCESS;

	if(g_mdb_stmt != NULL){
		SQLFreeHandle(SQL_HANDLE_STMT, g_mdb_stmt);
	}else{}

	if(g_mdb_dbc != NULL){
		SQLDisconnect(g_mdb_dbc);
		SQLFreeHandle(SQL_HANDLE_DBC, g_mdb_dbc);
	}else{}

	if(g_mdb_env != NULL){
		SQLFreeHandle(SQL_HANDLE_ENV, g_mdb_env);
	}else{}

	if(SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &g_mdb_env))){
		PRINT_WITH_TS(NGX_LOG_ERR, "Failed to alloc env handle\n");
	}else if(SQL_FAIL(retcode = SQLSetEnvAttr(g_mdb_env, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3_80, 0))
		|| SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_DBC, g_mdb_env, &g_mdb_dbc))){
		get_sql_diag(SQL_HANDLE_ENV, g_mdb_env, out_state, NULL);
		PRINT_WITH_TS(NGX_LOG_ERR, "%s\n", out_state);	
	}else if(SQL_FAIL(retcode = SQLConnect(g_mdb_dbc, (SQLCHAR*)(g_del_mdb_conn.dsn), SQL_NTS, 
												  	(SQLCHAR*)(g_del_mdb_conn.user), SQL_NTS, 
												  	(SQLCHAR*)(g_del_mdb_conn.passwd), SQL_NTS))
			|| SQL_FAIL(retcode = SQLSetConnectAttr(g_mdb_dbc, SQL_AUTOCOMMIT,
										(SQLPOINTER)SQL_AUTOCOMMIT_ON, SQL_IS_UINTEGER ) )
			|| SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_STMT, g_mdb_dbc, &g_mdb_stmt))){
		get_sql_diag(SQL_HANDLE_DBC, g_mdb_dbc, out_state, NULL);
		PRINT_WITH_TS(NGX_LOG_ERR, "%s\n", out_state);
	}else{}

	return SQL_SUCC(retcode);
}

extern int total_conn_num;
int g_rpt_total_conn_num = 0;
int g_seq_conn = 0;

bool report_conn_num()
{
    bool ret = true;

	if(pthread_mutex_lock(&g_del_mutex) == 0){
		if(g_rpt_total_conn_num != total_conn_num ||
			g_rpt_total_conn_num_del != g_total_conn_num_del){
			struct timeval tv;
			if(gettimeofday(&tv, NULL) == 0){
				if(tv.tv_sec > g_del_runtime.start + g_del_runtime.rup){
					char sql[DEF_SQL_LEN] = {'\0'};
					char out_state[DEF_MSG_LEN] = {'\0'};

					sprintf(sql, "INSERT INTO OB_CONN_LOG(start_time, g_id, n_id, pid, "
								"seq, rpt_time, conn_num_4, conn_num_del)"
								"VALUES('%s', %d, %d, %d, %d, %d, %d, %d);",
								g_del_start_str, g_del_server.gid, g_del_server.nid, 
								getpid(),
								g_seq_conn, (int)(tv.tv_sec - g_del_runtime.start), 
								total_conn_num, g_total_conn_num_del);

					if(SQL_FAIL(SQLExecDirect(g_mdb_stmt, (SQLCHAR*)sql, SQL_NTS))){
						get_sql_diag(SQL_HANDLE_STMT, g_mdb_stmt, out_state, NULL);
						PRINT_WITH_TS(NGX_LOG_ERR, "%s\nResult SQL:%s\n", out_state, sql);
						ret =  false;
					}else{
						g_rpt_conn_num_first_time = false;
						g_seq_conn ++;
						g_rpt_total_conn_num = total_conn_num;
						g_rpt_total_conn_num_del = g_total_conn_num_del;
					}					
				}else{}
			}else{
				ret =  false;
        		PRINT_WITH_TS(NGX_LOG_ERR, "gettimeofday failed errno:%d\n", errno);
			}
		}else{}

        if(pthread_mutex_unlock(&g_del_mutex) != 0){
            ret = false;
            PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno:%d\n", errno);
        }else{}
	}else{
        ret = false;
        PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
    }

    return ret;
}

bool report_summary_mdb()
{
	char sql[DEF_SQL_LEN] = {'\0'};
	char out_state[DEF_MSG_LEN] = {'\0'};
	SQLRETURN retcode = SQL_SUCCESS;

	sprintf(sql, "INSERT INTO TXN_STATS(start_time, g_id, n_id, deld_cmt_cnt, deld_lack_ord_cnt)"
				"VALUES('%s', %d, %d, %lld, %lld)"
				"ON DUPLICATE KEY UPDATE deld_cmt_cnt = ifnull(deld_cmt_cnt, 0) + %lld , "
										"deld_lack_ord_cnt = ifnull(deld_lack_ord_cnt, 0) + %lld;",
				g_del_start_str, g_del_server.gid, g_del_server.nid, g_del_stat_buf->cmt_num + g_del_stat_buf->lack_ord_num, 
				g_del_stat_buf->lack_ord_num,
				g_del_stat_buf->cmt_num + g_del_stat_buf->lack_ord_num, 
				g_del_stat_buf->lack_ord_num);

	if(SQL_FAIL(retcode = SQLExecDirect(g_mdb_stmt, (SQLCHAR*)sql, SQL_NTS))){
		get_sql_diag(SQL_HANDLE_STMT, g_mdb_stmt, out_state, NULL);
		PRINT_WITH_TS(NGX_LOG_ERR, "%s\nResult SQL:%s\n", out_state, sql);
	}else{
		PRINT_WITH_TS(NGX_LOG_INFO, "record delivery transactions statistics succeeds\n");
	}

	return SQL_SUCC(retcode);
}

extern txn_stat all_tpcc_txns[TXN_TYPE_NUM_PLUS];  
static int g_total_del_task_processed = 0;

bool record_realtime_txn_processor(long long int finish_time, int dtt, bool succ, int qt, int rt)
{
	bool ret = true;

	if(pthread_mutex_lock(&g_del_mutex) == 0){
		int idx = (finish_time + dtt - g_del_runtime.start) / g_del_runtime.ri;	
		
		if(idx >= 0 && idx < g_del_rts_buf->point_num){
			g_del_rts_buf->rts[idx].cmt_num += succ;
			g_del_rts_buf->rts[idx].err_num += !succ;
		}else{}

		if(succ){
			all_tpcc_txns[DELIVERY_DEF].txn_cnt ++;
			all_tpcc_txns[DELIVERY_DEF].total_in_time += qt;
			all_tpcc_txns[DELIVERY_DEF].total_db_time += rt;
			if(all_tpcc_txns[DELIVERY_DEF].max_in_time < qt){
				all_tpcc_txns[DELIVERY_DEF].max_in_time = qt;
			}else{}

			if(all_tpcc_txns[DELIVERY_DEF].max_db_time < rt){
				all_tpcc_txns[DELIVERY_DEF].max_db_time = rt;
			}else{}
		}else{}

		if(pthread_mutex_unlock(&g_del_mutex) != 0){
			ret = false;
			PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno:%d\n", errno);
		}else{}
		
	}else{
		ret = false;
		PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
	}
	
	return ret;
}

bool record_realtime_txn_dumper(long long int cur_time)
{
	bool ret = true;
	SQLRETURN retcode = SQL_SUCCESS;
	bool bfirst = true;

	if(pthread_mutex_lock(&g_del_mutex) == 0){
		bfirst = g_rpt_conn_num_first_time;
	
		int cur_idx = (cur_time - g_del_runtime.start) / g_del_runtime.ri;
		
		if((g_del_last_rpt_idx + RPT_DELAY_NUM < cur_idx) 
			&& (g_del_last_rpt_idx < g_del_rts_buf->point_num)){
			RealtimeStat * rts = (RealtimeStat *)&(g_del_rts_buf->rts);
			char sql[DEF_SQL_LEN] = {'\0'};
			char msg[DEF_MSG_LEN] = {'\0'};

			cur_idx -= RPT_DELAY_NUM;
			while(cur_idx > g_del_last_rpt_idx){
				g_del_last_rpt_idx ++;
				sprintf(sql, "INSERT INTO RTS_V2(start_time, n_id, g_id, seq, "
									"del_def_cmt_num, del_def_err_num) "
									"VALUES('%s', %d, %d, %d, %lld, %lld) "
									"ON DUPLICATE KEY UPDATE del_def_cmt_num = ifnull(del_def_cmt_num, 0) + %lld, "
															"del_def_err_num = ifnull(del_def_err_num, 0) + %lld;",
									g_del_start_str, 
									g_del_server.nid, 
									g_del_server.gid, 
									g_del_last_rpt_idx, 
									rts[g_del_last_rpt_idx].cmt_num, 
									rts[g_del_last_rpt_idx].err_num,
									rts[g_del_last_rpt_idx].cmt_num,
									rts[g_del_last_rpt_idx].err_num);
						
				if(SQL_FAIL(retcode = SQLExecDirect(g_mdb_stmt, (SQLCHAR*)sql, SQL_NTS))){
					get_sql_diag(SQL_HANDLE_STMT, g_mdb_stmt, msg, NULL);
					PRINT_WITH_TS(NGX_LOG_ERR, "SQLExecDirect failed SQLRETURN: %d %s\nSQL:%s\n", retcode, msg, sql);
				}else{}
			} //end while
		}else{}

        if(pthread_mutex_unlock(&g_del_mutex) != 0){
            ret = false;
            PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno:%d\n", errno);
        }else{}
	}else{
		ret = false;
		PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
	}

	// report connection num 
	if(ret && (cur_time >= g_del_runtime.start + g_del_runtime.rup) &&
		bfirst && !report_conn_num()){
		ret = false;
		PRINT_WITH_TS(NGX_LOG_ERR, "report_conn_num failed\n");
	}else{}

	return ret;
}

 
bool record_delivery_stat( int w_id, 
						int o_carrier_id, 
						int o_cnt, 
						int* d_ids,		
						int* o_ids, 
						long long int enqueue_time, 
						long long int dequeue_time, 
						long long int finish_time,
						bool succ)
{
	bool ret = true;
	
	if(pthread_mutex_lock(&g_del_mutex) == 0){
		while(ret && (g_del_stat_buf->head - g_del_stat_buf->tail >= g_del_stat_buf->total_num)){
			PRINT_WITH_TS(NGX_LOG_WARN, "delivery stat buffer is full\n");
			if(!(ret = (pthread_mutex_unlock(&g_del_mutex) == 0))){
				PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno:%d\n", errno);
			}else if(!(ret = delivery_dumper(false))){
			}else if(!(ret = (pthread_mutex_lock(&g_del_mutex) == 0))){
			}else{}
		}
		
		if(ret){
			if(succ){
				long long int head = g_del_stat_buf->head % g_del_stat_buf->total_num;
				int rt = 0, qt = 0, rt_ms = 0, qt_ms = 0;

				DeliveryStat * del_stat = &g_del_stat_buf->del_stat[head];

				del_stat->w_id = w_id;
				del_stat->o_carrier_id = o_carrier_id;
				del_stat->o_cnt = o_cnt;
				memcpy(del_stat->d_ids, d_ids, o_cnt * sizeof(int));
				memcpy(del_stat->o_ids, o_ids, o_cnt * sizeof(int));
				del_stat->enqueue_time = enqueue_time;
				del_stat->dequeue_time = dequeue_time;
				del_stat->finish_time = finish_time;

				qt = dequeue_time - enqueue_time;
				rt = finish_time - enqueue_time;

				g_del_rt[head] = rt;

				rt_ms = MT_2_MSEC_CEIL(rt);
				qt_ms = MT_2_MSEC_CEIL(qt);
				g_del_stat_buf->total_rt += rt_ms;
				g_del_stat_buf->total_qt += qt_ms;
				g_del_stat_buf->min_rt = g_del_stat_buf->min_rt > rt_ms ? rt_ms : g_del_stat_buf->min_rt;
				g_del_stat_buf->max_rt = g_del_stat_buf->max_rt < rt_ms ? rt_ms : g_del_stat_buf->max_rt;
				
				g_del_stat_buf->head ++;

				if(o_cnt < DISTS_PER_WARE){
					g_del_stat_buf->lack_ord_num ++;
					PRINT_WITH_TS(NGX_LOG_ERR, "delivery order cnt unexpected: w_id: %d o_cnt: %d\n", w_id, o_cnt);
				}else{
					g_del_stat_buf->cmt_num ++;
				}
			}else{
				g_del_stat_buf->err_num ++;
			}

			if(pthread_mutex_unlock(&g_del_mutex) != 0){
				ret = false;
				PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno:%d\n", errno);
			}else{}
		}else{}
	}else{
		ret = false;
		PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
	}

	return ret;
}

static bool reconnect_ob_servers()
{
	SQLRETURN retcode = SQL_SUCCESS;
	bool ret = true;
	char out_state[DEF_MSG_LEN] ={'\0'};
	bool need_reconn = false;
	int  conn_chg = 0;

	if(g_del_stmt){
		SQLFreeHandle(SQL_HANDLE_STMT, g_del_stmt);
		g_del_stmt = NULL;
	}

	if(g_del_dbc){
		SQLDisconnect(g_del_dbc);
		SQLFreeHandle(SQL_HANDLE_DBC, g_del_dbc);
		g_del_dbc = NULL;
		conn_chg -= 1;
	}

	if(g_del_env){
		SQLFreeHandle(SQL_HANDLE_ENV, g_del_env);
		g_del_env = NULL;
	}

	if(SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &g_del_env))){
        PRINT_WITH_TS(NGX_LOG_ERR, "Failed to alloc g_del_env handle\n");
        ret = false;
    }else if(SQL_FAIL(retcode = SQLSetEnvAttr(g_del_env, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3_80, 0))
        || SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_DBC, g_del_env, &g_del_dbc))){
        get_sql_diag(SQL_HANDLE_ENV, g_del_env, out_state, NULL);
        PRINT_WITH_TS(NGX_LOG_ERR, "%s\n", out_state);
        ret = false;
    }else if( SQL_FAIL( retcode = SQLConnect(g_del_dbc, (SQLCHAR*)(g_del_conn.dsn), SQL_NTS, (SQLCHAR*)(g_del_conn.user),
                                            SQL_NTS, (SQLCHAR*)(g_del_conn.passwd), SQL_NTS))
    		|| SQL_FAIL( retcode = SQLSetConnectAttr(g_del_dbc, SQL_ATTR_TXN_ISOLATION,
                                        (SQLPOINTER)SQL_TRANSACTION_SERIALIZABLE, SQL_IS_UINTEGER ) )
    		|| SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_STMT, g_del_dbc, &g_del_stmt))){	
		get_sql_diag(SQL_HANDLE_DBC, g_del_dbc, out_state, NULL);
		PRINT_WITH_TS(NGX_LOG_ERR, "%s\n", out_state);
		ret = false;
	}else if(SQL_FAIL(retcode = SQLPrepare(g_del_stmt, (SQLCHAR*)DELIVERY_PROC, SQL_NTS))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
					   0, 0, &del_w_id, 0, NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 2, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
					   SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &del_d_ids_arr, sizeof(ARRAY_STRUCT), NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 3, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
					   SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &del_o_ids_arr, sizeof(ARRAY_STRUCT), NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 4, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
					   0, 0, &del_o_cnt, 0, NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 5, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
					   SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &del_sums_arr, sizeof(ARRAY_STRUCT), NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 6, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR,
					   30, 0, del_d_date, 30, NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 7, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
					   0, 0, &del_o_carrier_id, 0, NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 8, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
					   SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &del_c_ids_arr, sizeof(ARRAY_STRUCT), NULL))
			|| SQL_FAIL(retcode = SQLBindParameter(g_del_stmt, 9, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
					   0, 0, &del_retry, 0, NULL))){
		get_sql_diag(SQL_HANDLE_STMT, g_del_stmt, out_state, &need_reconn);
		PRINT_WITH_TS(NGX_LOG_ERR, "%s\n", out_state);
		ret = false;
	}else{
		conn_chg += 1;
	}

	if(conn_chg){
		if(pthread_mutex_lock(&g_del_mutex) == 0){
			g_total_conn_num_del += conn_chg;
			if(pthread_mutex_unlock(&g_del_mutex) != 0){
				PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno: %d\n", errno);
			}else{
				report_conn_num();
			}
		}else{
			PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
		}
	}else{}

	return ret;
}

bool prepare_del_params(DeliveryTask * del_task)
{
	del_w_id = del_task->w_id;
	del_o_carrier_id= del_task->o_carrier_id;

	g_total_del_task_processed ++;	
	
	return true;
}

void delivery_processor(void *del_data, ngx_log_t *log)
{
	long long int enqueue_time = 0, dequeue_time = 0, finish_time = 0;

	SQLRETURN retcode = SQL_SUCCESS;
	char out_state[DEF_MSG_LEN] ={'\0'};

	if(!g_del_thread_inited){
		int i = 0;

		for(i = 0; i < DISTS_PER_WARE; i ++){
			del_d_ids_num[i].type = SQL_C_LONG;
			del_d_ids_num[i].is_null = false;
			del_d_ids_num[i].data_buffer = &(del_d_ids[i]);
			del_o_ids_num[i].type = SQL_C_LONG;
			del_o_ids_num[i].is_null = false;
			del_o_ids_num[i].data_buffer = &(del_o_ids[i]);
			del_sums_num[i].type = SQL_C_LONG;
			del_sums_num[i].is_null = false;
			del_sums_num[i].data_buffer = &(del_sums[i]);
			del_c_ids_num[i].type = SQL_C_LONG;
			del_c_ids_num[i].is_null = false;
			del_c_ids_num[i].data_buffer = &(del_c_ids[i]);
		}

		del_d_ids_arr.type = SQL_DRIVER_C_ARRAY;
		del_d_ids_arr.type_name = NULL;
		del_d_ids_arr.size = DISTS_PER_WARE;
		del_d_ids_arr.data_buffer = (void *)del_d_ids_num;
		del_o_ids_arr.type = SQL_DRIVER_C_ARRAY;
		del_o_ids_arr.type_name = NULL;
		del_o_ids_arr.size = DISTS_PER_WARE;
		del_o_ids_arr.data_buffer = (void *)del_o_ids_num;
		del_sums_arr.type = SQL_DRIVER_C_ARRAY;
		del_sums_arr.type_name = NULL;
		del_sums_arr.size = DISTS_PER_WARE;
		del_sums_arr.data_buffer = (void *)del_sums_num;
		del_c_ids_arr.type = SQL_DRIVER_C_ARRAY;
		del_c_ids_arr.type_name = NULL;
		del_c_ids_arr.size = DISTS_PER_WARE;
		del_c_ids_arr.data_buffer = (void *)del_c_ids_num;

		g_del_thread_inited = true;
	}else{}

	if( (!g_del_env || !g_del_dbc || !g_del_stmt) && !reconnect_ob_servers()){
		PRINT_WITH_TS(NGX_LOG_ERR, "obdelivery: failed to connect to OceanBase\n");
	}else{
		long long int test_end = (g_del_runtime.start + g_del_runtime.rup +
							  	g_del_runtime.rdown + g_del_runtime.mi);
		long long int mi_begin = g_del_runtime.start + g_del_runtime.rup;
		long long int mi_end = mi_begin + g_del_runtime.mi;
		int dtt = 0;

		if(!get_timestamp(del_d_date)){
			PRINT_WITH_TS(NGX_LOG_ERR, "failed to get_timestamp\n");
		}else{ 
			DeliveryTask * del_task = (DeliveryTask *)del_data; 	
			struct timeval dequeue_tv, finish_tv;
	
			if(gettimeofday(&dequeue_tv, NULL) < 0){
				PRINT_WITH_TS(NGX_LOG_ERR, "failed to gettimeofday\n");
			}else{
				bool need_reconn = false;
				
				prepare_del_params(del_task);

				dtt = del_task->dtt;
				enqueue_time= del_task->enqueue_time;

				if(SQL_FAIL(retcode = SQLExecute(g_del_stmt))){
					char msg[DEF_MSG_LEN] = {'\0'};

					get_sql_diag(SQL_HANDLE_STMT, g_del_stmt, out_state, &need_reconn);
					PRINT_WITH_TS(NGX_LOG_ERR, "delivery failed %s\n", out_state);

					if(!need_reconn){
						SQLFreeStmt(g_del_stmt, SQL_CLOSE);
						if(!rollback_on_error(g_del_dbc, msg, &need_reconn)){
							PRINT_WITH_TS(NGX_LOG_ERR, "%s\n", msg);
						}else{}
					}else{
						bool reconn = false;
						bool reach_test_end = false;

						while(!reconn && !reach_test_end){
							struct timeval cur_tv;
					
							sleep(DEL_RECONN_WAIT); // wait to reconnect
							reconn = reconnect_ob_servers(g_del_conn);
							PRINT_WITH_TS(NGX_LOG_ERR, "delivery reconnect to oceanbase servers %s\n", 
														reconn ? "succeed" : "fail");

							if(gettimeofday(&cur_tv, NULL) < 0){
								PRINT_WITH_TS(NGX_LOG_ERR, "failed to gettimeofday\n");
							}else{
								reach_test_end = (cur_tv.tv_sec >= test_end);
							}
						}
					}
				}else{}

				if(gettimeofday(&finish_tv, NULL) < 0){
					PRINT_WITH_TS(NGX_LOG_ERR, "failed to gettimeofday\n");
				}else{
					dequeue_time = TV_2_MT(dequeue_tv); 
					finish_time = TV_2_MT(finish_tv); 

					if((MT_2_SEC(dequeue_time) >= mi_begin && MT_2_SEC(dequeue_time) < mi_end)
						&& (MT_2_SEC(finish_time) >= mi_begin && MT_2_SEC(finish_time) < mi_end)){
						record_delivery_stat(del_w_id, 
											del_o_carrier_id, 
											del_o_cnt,
											del_d_ids, 
											del_o_ids, 
											enqueue_time, 
											dequeue_time,
											finish_time,
											SQL_SUCC(retcode));
					}else{}

					if(!record_realtime_txn_processor(MT_2_SEC(finish_time),
											dtt,
											SQL_SUCC(retcode),
											MT_2_MSEC_CEIL(dequeue_time - enqueue_time),
											MT_2_MSEC_CEIL(finish_time - dequeue_time))){
						PRINT_WITH_TS(NGX_LOG_ERR, "failed to record_realtime_txn\n");
					}else{}
				}//end else
			}//end else
		}//end else
	}

	return ;
}

bool create_result_files()
{
	char res_file[128] = {'\0'};
	char start_str[32] = {'\0'};
	char * prefix[DEL_FILES_NUM] = {"del_resf", "del_def"};
	FILE * res = NULL;
	bool ret = true;

	if(!get_start_time_str(g_del_runtime.start, start_str)){
		ret = false;		
	}else{
		int i = 0;

		for(i = 0; (i < DEL_FILES_NUM) && ret; i ++){
			sprintf(res_file, "out/%s_%d_%s_%d",
					   prefix[i],
					   g_del_server.nid,
					   start_str,
					   getpid());

			if(!(res = fopen(res_file, "w+"))){
				ret = false;
				PRINT_WITH_TS(NGX_LOG_ERR, "fopen %s returns NULL errno: %d\n", res_file, errno);
			}else{
				g_del_res_files[i] = res;
			}
		}
    }

	return ret;
}

bool delivery_dumper(bool all)
{
	bool succ = true;
	FILE * fp_drf = g_del_res_files[DEL_RES_FILE];
	FILE * fp_rt  = g_del_res_files[DEL_RT_FILE];

	if(fp_drf && fp_rt){ 
		if(pthread_mutex_lock(&g_del_mutex) == 0){
			while((g_del_stat_buf->head - g_del_stat_buf->tail >= WB_BATCH_SIZE) && succ){
				int start_pos = g_del_stat_buf->tail % g_del_stat_buf->total_num;

				if(start_pos + WB_BATCH_SIZE <= g_del_stat_buf->total_num){
					if((fwrite(&(g_del_stat_buf->del_stat[start_pos]), 
								sizeof(DeliveryStat), WB_BATCH_SIZE, fp_drf) < WB_BATCH_SIZE)
						|| (fwrite(g_del_rt + start_pos, sizeof(int), WB_BATCH_SIZE, fp_rt) <  WB_BATCH_SIZE)){
						succ = false;
					}else{
						fflush(fp_drf);
						fflush(fp_rt);
					}
				}else{
					succ = false;
				}
				
				g_del_stat_buf->tail += WB_BATCH_SIZE;
			}//end while
			
			if(pthread_mutex_unlock(&g_del_mutex) != 0){
				PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno:%d\n", errno);
				succ = false;
			}else{}
		}else{
			succ = false;
			PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
		}// end if

		// dump all to file before exit
		if(all && succ){
			if(pthread_mutex_lock(&g_del_mutex) == 0){
				unsigned int cnt = g_del_stat_buf->head - g_del_stat_buf->tail;
				int start_pos = g_del_stat_buf->tail % g_del_stat_buf->total_num;
				
				if(cnt > 0){	
					if(start_pos + cnt <= (unsigned)g_del_stat_buf->total_num){	
						if((fwrite(&(g_del_stat_buf->del_stat[start_pos]), sizeof(DeliveryStat), cnt, fp_drf) < (unsigned int)cnt)
							|| (fwrite(g_del_rt + start_pos, sizeof(int), cnt, fp_rt) < (unsigned int)cnt)){
							succ = false;
							PRINT_WITH_TS(NGX_LOG_ERR, "fwrite failed errno: %d\n", errno);
						}else{}
					}else{
						unsigned int first_part = g_del_stat_buf->total_num - start_pos;

						if((fwrite(&(g_del_stat_buf->del_stat[start_pos]), sizeof(DeliveryStat), first_part, fp_drf) < first_part)
						|| (fwrite(&(g_del_stat_buf->del_stat[0]), sizeof(DeliveryStat), cnt - first_part, fp_drf) 
																									< (cnt - first_part))
						|| (fwrite(g_del_rt + start_pos, sizeof(int), first_part, fp_rt) < first_part)
						|| (fwrite(g_del_rt, sizeof(int), cnt - first_part, fp_rt) < (cnt - first_part))){
							succ = false;
							PRINT_WITH_TS(NGX_LOG_ERR, "fwrite failed errno: %d\n", errno);
						}else{}		
					}

					g_del_stat_buf->tail = g_del_stat_buf->head;

					fflush(fp_drf);
					fflush(fp_rt);
				}else{}

				//record to metadb
				report_summary_mdb();

				if(pthread_mutex_unlock(&g_del_mutex) != 0){
					succ = false;
					PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_unlock failed errno: %d\n", errno);
				}else{}
			}else{
				succ = false;
				PRINT_WITH_TS(NGX_LOG_ERR, "pthread_mutex_lock failed errno:%d\n", errno);
			}
		}else{}
	}else{
		PRINT_WITH_TS(NGX_LOG_ERR, "result files is not initialized\n");
	}

	return succ;
}

void delivery_dumper_main(void *del_data, ngx_log_t *log)
{
	long long int test_end = g_del_runtime.start + g_del_runtime.rup + g_del_runtime.rdown + g_del_runtime.mi 
							+ DEL_DUMPER_DELAY;
	bool bSucc = true, bEnd = false;	
	struct timeval tv_cur;

	for(; bSucc && !bEnd; ){
		if(!(bSucc = delivery_dumper(false))){
			PRINT_WITH_TS(NGX_LOG_ERR, "delivery_dumper(false) failed\n");
		}else if(gettimeofday(&tv_cur, 0) != 0){
			bSucc = false;
			PRINT_WITH_TS(NGX_LOG_ERR, "gettimeofday failed errno: %d\n", errno);
		}else if(!(bSucc = record_realtime_txn_dumper(TV_2_SEC(tv_cur)))){
			PRINT_WITH_TS(NGX_LOG_ERR, "report_realtime_txn_dumper failed\n");
		}else if(TV_2_SEC(tv_cur) > test_end){
			bEnd = true;
		}else{
			sleep(DEL_DUMPER_WAIT);
		}
	}

	if(bSucc && bEnd){
		if(!(bSucc = delivery_dumper(true))){
			PRINT_WITH_TS(NGX_LOG_ERR, "delivery_dumper(true) failed\n");
		}else{
			PRINT_WITH_TS(NGX_LOG_INFO, "delivery dumper thread finished\n");
		} 
	}
}
