/*
 * ngx_tpcc_delivery.h is delivery daemon specified declaration
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */
#ifndef __TPCC_DELIVERY_DAEMON_H__
#define __TPCC_DELIVERY_DAEMON_H__

#include <pthread.h>
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>
#include "ngx_tpcc_common.h"

#define RPT_DELAY_NUM 2
#define DEL_DUMPER_DELAY 120
#define DEF_DEL_TASK_IN_QUEUE 100000
#define DISTS_PER_WARE 10
#define WB_BATCH_SIZE 10000
#define MAX_DEL_STAT_BUF_SIZE (100 * WB_BATCH_SIZE)

#define PNT_NUM_PLUS 600
#define DEL_RECONN_WAIT		5
#define DEL_DUMPER_WAIT		1
#define DEF_MIN_RT		10000

#define DUMPER_W_ID -1

#define	DEL_RES_FILE	0
#define DEL_RT_FILE		1
#define DEL_FILES_NUM	2

#define PRINT_WITH_TS(log_level, fmt, ...) do {\
                                    char msg[DEF_MSG_LEN] ={'\0'}; \
                                    snprintf(msg, DEF_MSG_LEN, "%s "fmt, get_cur_time(), ##__VA_ARGS__);\
                                    ngx_log_error(log_level, ((ngx_cycle_t *) ngx_cycle)->log, 0, "%s\n", msg);\
                                }while(0)

typedef	struct _delivery_stat{
    long long int enqueue_time;
    long long int dequeue_time;
    long long int finish_time;
    int w_id;
    int o_carrier_id;
    int o_cnt;
    int d_ids[DISTS_PER_WARE];
    int o_ids[DISTS_PER_WARE];
}DeliveryStat;

typedef struct _delivery_stat_buf{
    pthread_mutex_t * mutex;
	long long int cmt_num;
	long long int err_num;
	long long int lack_ord_num;
	long long int total_qt;
	long long int total_rt;
	int max_rt;
	int min_rt;
    int total_num;
    volatile long long head;
    volatile long long tail;
    DeliveryStat del_stat[1];
}DeliveryStatBuf;

typedef struct{
	int gid;
	int nid;
	int qsize;
}DeliveryServer;

typedef struct {
    int start;
    int rup;
    int rdown;
    int mi;
    int ri;
    int wts;
}RuntimeConf;

typedef struct _realtime_stat{
	long long int cmt_num;
	long long int err_num;
}RealtimeStat;

typedef struct _realtime_stat_buf{
	int point_num;
	RealtimeStat rts[1];
}RealtimeStatBuf;

extern bool get_start_time_str(long start, char* start_str);
extern pthread_mutex_t  g_del_mutex;
extern FILE * g_del_res_files[2];
extern DeliveryStatBuf * g_del_stat_buf;
extern int * g_del_rt;
extern RealtimeStatBuf * g_del_rts_buf;

extern DeliveryServer g_del_server;
extern ConnInfo g_del_mdb_conn;
extern ConnInfo g_del_conn;
extern RuntimeConf g_del_runtime;
extern char g_del_start_str[30];

extern SQLHANDLE g_mdb_env;
extern SQLHANDLE g_mdb_dbc;
extern SQLHANDLE g_mdb_stmt;

extern bool init_del_statbuf();
extern bool init_del_rts_points(int rup, int rdown, int mi, int ri);
extern bool init_metadb();
extern bool create_result_files();
extern void delivery_processor(void *del_data, ngx_log_t *log);
extern bool report_conn_num();
extern void delivery_dumper_main(void *del_data, ngx_log_t *log);
#endif
