/*
 * ngx_tpcc_txn.c is tpcc txn processing routine
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/time.h>
#include <ngx_core.h>
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>
#include "ngx_tpcc_common.h"
#include "ngx_tpcc_txn.h"
#include "ngx_tpcc_delivery_thread.h"
#include "ngx_tpcc_delivery_pool.h"
#include "ngx_http_oceanbase_output.h"

#include <unistd.h>


txn_stat all_tpcc_txns[TXN_TYPE_NUM_PLUS] = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}};
int last_report_time = 0;
bool gen_dumper_thread = false;

extern long long int g_max_txn_cnt;

bool put_dumper_task(ngx_log_t * log)
{
	bool bSucc = false;
	ngx_thread_task_t * del_task =  NULL;

	if(gen_dumper_thread){
		bSucc = true;
	}else if(pthread_mutex_init(&g_del_mutex, NULL) != 0){
        ngx_log_error(NGX_LOG_ERR, log, 0, "pthread_mutex_init failed");
	}else if(!(bSucc = get_start_time_str(g_del_runtime.start, g_del_start_str))){
		ngx_log_error(NGX_LOG_ERR, log, 0, "get_start_time_str failed");
	}else if(!(bSucc = init_del_statbuf())){
		ngx_log_error(NGX_LOG_ERR, log, 0, "init_del_statbuf failed");
	}else if(!(bSucc = init_del_rts_points(g_del_runtime.rup,
											g_del_runtime.rdown,
											g_del_runtime.mi,
											g_del_runtime.ri))){
		ngx_log_error(NGX_LOG_ERR, log, 0, "init_del_rts_points failed");
	}else if(!(bSucc = init_metadb())){
		ngx_log_error(NGX_LOG_ERR, log, 0, "init_metadb failed");
	}else if(!(bSucc = create_result_files())){
		ngx_log_error(NGX_LOG_ERR, log, 0, "create_result_files failed");
	}else if(!(del_task = get_del_task_from_pool())){
       	ngx_log_error(NGX_LOG_ERR, log, 0, "fail to get_del_task_from pool");
	}else{ 
		DeliveryTask * del_task_inner = (DeliveryTask *)(del_task->ctx);
 
	    del_task->handler = delivery_dumper_main;
    	del_task->event.handler = del_thread_completion;
    	del_task->event.data = del_task_inner;
    	del_task->event.log  = log;

		del_task_inner->w_id = DUMPER_W_ID;
		del_task_inner->o_carrier_id = 0;
		del_task_inner->dtt = 0;
		del_task_inner->enqueue_time = 0;
		
		if (ngx_thread_task_post(g_del_thread_pool, del_task) != NGX_OK) {
        	ngx_log_error(NGX_LOG_ERR, log, 0, "fail to do ngx_thread_task_post");
    	}else{
			gen_dumper_thread = true;
			bSucc = true;
        	ngx_log_error(NGX_LOG_INFO, log, 0, "ngx_thread_task_post succeeds");
		}
	}

	return bSucc;
}

bool put_delivery_task(int w_id, int o_carrier_id, int dtt, long long int rte_send_ts, char* status, ngx_log_t *log)
{
	bool bSucc = true;
	struct timeval enqueue_time;
	
	ngx_thread_task_t * del_task =  NULL;

	if(gettimeofday(&enqueue_time, 0) < 0){
		bSucc = false;
	}else if(!gen_dumper_thread && !put_dumper_task(log)){
       	ngx_log_error(NGX_LOG_ERR, log, 0, "fail to put_dumper_task");
		bSucc = false;
	}else if(!(del_task = get_del_task_from_pool())){
       	ngx_log_error(NGX_LOG_ERR, log, 0, "fail to get_del_task_from pool");
		bSucc = false;
	}else{
		int cur_time = enqueue_time.tv_sec;
		int in_time = (int)(TV_2_MSEC(enqueue_time) - rte_send_ts);
		DeliveryTask * del_task_inner = (DeliveryTask *)(del_task->ctx);
 
	    del_task->handler = delivery_processor;
    	del_task->event.handler = del_thread_completion;
    	del_task->event.data = del_task_inner;
    	del_task->event.log  = log;

		del_task_inner->w_id = w_id;
		del_task_inner->o_carrier_id = o_carrier_id;
		del_task_inner->dtt = dtt;
		del_task_inner->enqueue_time =
							enqueue_time.tv_sec * 10000LL + enqueue_time.tv_usec /100; // 1/10 ms
		
		if (ngx_thread_task_post(g_del_thread_pool, del_task) != NGX_OK) {
        	ngx_log_error(NGX_LOG_ERR, log, 0, "fail to do ngx_thread_task_post");
			bSucc = false;
    	}

		all_tpcc_txns[DELIVERY].txn_cnt ++;
		all_tpcc_txns[DELIVERY].total_in_time += in_time;
		if(in_time > all_tpcc_txns[DELIVERY].max_in_time){
			all_tpcc_txns[DELIVERY].max_in_time = in_time;
		}else{}

		if(status && bSucc){
			strcpy(status, DEILIVERY_QUEUED_MSG);
		}else{}

		if(cur_time - last_report_time > REPORT_TXN_INTERVAL){
			char log_txt[1024] = {0};
			last_report_time = cur_time;
			sprintf(log_txt, "Worker %d transaction statistics:\n"
							 "type:  no       |   pay   | del-ia/del-def |   os    |   sl    \n"
							 "num:   %9lld|%9lld|%9lld/%9lld|%9lld|%9lld\n"
							 "avg-rte(ms): %4.1f|%4.1f|%4.1f|%4.1f|%4.1f\n"
							 "avg-db(ms):  %4.1f|%4.1f|%4.1f/%4.1f|%4.1f|%4.1f\n"
							 "max-rte(ms): %4d|%4d|%4d|%4d|%4d\n"
							 "max-db(ms):  %4d|%4d|%4d/%4d|%4d|%4d\n",
									getpid(),
									all_tpcc_txns[NEW_ORDER].txn_cnt,
									all_tpcc_txns[PAYMENT].txn_cnt,
									all_tpcc_txns[DELIVERY].txn_cnt,
									all_tpcc_txns[DELIVERY_DEF].txn_cnt,
									all_tpcc_txns[ORDER_STATUS].txn_cnt,
									all_tpcc_txns[STOCK_LEVEL].txn_cnt,
all_tpcc_txns[NEW_ORDER].txn_cnt ? all_tpcc_txns[NEW_ORDER].total_in_time*1.0/all_tpcc_txns[NEW_ORDER].txn_cnt : 0,
all_tpcc_txns[PAYMENT].txn_cnt ? all_tpcc_txns[PAYMENT].total_in_time*1.0/all_tpcc_txns[PAYMENT].txn_cnt : 0,
all_tpcc_txns[DELIVERY].txn_cnt ? all_tpcc_txns[DELIVERY].total_in_time*1.0/all_tpcc_txns[DELIVERY].txn_cnt : 0,
all_tpcc_txns[ORDER_STATUS].txn_cnt ? all_tpcc_txns[ORDER_STATUS].total_in_time*1.0/all_tpcc_txns[ORDER_STATUS].txn_cnt : 0,
all_tpcc_txns[STOCK_LEVEL].txn_cnt ? all_tpcc_txns[STOCK_LEVEL].total_in_time*1.0/all_tpcc_txns[STOCK_LEVEL].txn_cnt : 0,
all_tpcc_txns[NEW_ORDER].txn_cnt ? all_tpcc_txns[NEW_ORDER].total_db_time*1.0/all_tpcc_txns[NEW_ORDER].txn_cnt:0,
all_tpcc_txns[PAYMENT].txn_cnt ? all_tpcc_txns[PAYMENT].total_db_time*1.0/all_tpcc_txns[PAYMENT].txn_cnt:0,
all_tpcc_txns[DELIVERY_DEF].txn_cnt ? all_tpcc_txns[DELIVERY_DEF].total_in_time*1.0/all_tpcc_txns[DELIVERY_DEF].txn_cnt:0,
all_tpcc_txns[DELIVERY_DEF].txn_cnt ? all_tpcc_txns[DELIVERY_DEF].total_db_time*1.0/all_tpcc_txns[DELIVERY_DEF].txn_cnt:0,
all_tpcc_txns[ORDER_STATUS].txn_cnt ? all_tpcc_txns[ORDER_STATUS].total_db_time*1.0/all_tpcc_txns[ORDER_STATUS].txn_cnt:0,
all_tpcc_txns[STOCK_LEVEL].txn_cnt ? all_tpcc_txns[STOCK_LEVEL].total_db_time*1.0/all_tpcc_txns[STOCK_LEVEL].txn_cnt:0,
				all_tpcc_txns[NEW_ORDER].max_in_time,
				all_tpcc_txns[PAYMENT].max_in_time,
				all_tpcc_txns[DELIVERY].max_in_time,
				all_tpcc_txns[ORDER_STATUS].max_in_time,
				all_tpcc_txns[STOCK_LEVEL].max_in_time,
				all_tpcc_txns[NEW_ORDER].max_db_time,
				all_tpcc_txns[PAYMENT].max_db_time,
				all_tpcc_txns[DELIVERY_DEF].max_in_time,
				all_tpcc_txns[DELIVERY_DEF].max_db_time,
				all_tpcc_txns[ORDER_STATUS].max_db_time,
				all_tpcc_txns[STOCK_LEVEL].max_db_time);
			ngx_log_error(NGX_LOG_ALERT, log, 0, "%s", log_txt);
		}else{}
	}
	
	return bSucc;
}

extern TERMINAL_STATES* g_terminal_states;

bool check_set_terminal_state(int t_id, int set,  ngx_log_t *log)
{
	return true;

	bool bSucc = true;

	if(!g_terminal_states){
		bSucc = false;
        ngx_log_error(NGX_LOG_ERR, log, 0, "check_set_terminal_state failed, g_terminal_states not initialized\n");
	}else{
		TERMINAL_STATE * term = &(g_terminal_states->terminal_state[t_id % g_terminal_states->max_terminal_cnt]);
		if(!term->terminal_id){
			term->terminal_id = t_id;
			term->active = set;
		}else if(term->terminal_id == t_id){
			if(set == !term->active){
				term->active = set;
			}else{
				bSucc = false;
        		ngx_log_error(NGX_LOG_ERR, log, 0, 
								"check_set_terminal_state failed, terminal %d state[%d %d] conflicts\n",
								t_id, term->active, set);
			}
		}else{
			bSucc = false;
        	ngx_log_error(NGX_LOG_ERR, log, 0, 
				"check_set_terminal_state failed, terminal [%d %d] unexpected\n",
				term->terminal_id, t_id );
		}
	}

	return bSucc ;
}

int trans_html_string(char* dst, char* src, int ilen)
{
	int	len = 0;
	
	while(*src){
		switch(*src){
		case '>' :
			*dst ++ = '&';
			*dst ++ = 'g';
			*dst ++ = 't';
			*dst ++ = ';';
			len += 4;
			ilen -= 4;
			break;
		case '<' :
			*dst ++ = '&';
			*dst ++ = 'l';
			*dst ++ = 't';
			*dst ++ = ';';
			len += 4;
			ilen -= 4;
			break;
		case '\"':
			*dst ++ = '&';
			*dst ++ = 'q';
			*dst ++ = 'u';
			*dst ++ = 'o';
			*dst ++ = 't';
			len += 5;
			ilen -= 5;
			break;
		case '&':
			*dst ++ = '&';
			*dst ++ = 'a';
			*dst ++ = 'm';
			*dst ++ = 'p';
			len += 4;
			ilen -= 4;
			break;
		default:
			*dst++ = *src;
			len += 1;
			ilen -= 1;
			break;
		}

		src ++;
	}

	while(ilen-- > 0){
		*dst ++ = ' ';
		len ++;
	}
	
	return len;
}

static int item_cmp(const void *i1, const void *i2)
{
	int ret  = 0;
	assert(i1 != NULL && i2 != NULL);
	if( (((ItemInfo *)i1)->ol_i_id < ((ItemInfo *)i2)->ol_i_id) || 
		(((ItemInfo *)i1)->ol_s_w_id < ((ItemInfo *)i2)->ol_s_w_id &&
		 ((ItemInfo *)i1)->ol_i_id == ((ItemInfo *)i2)->ol_i_id) ){
		ret = -1;
	}else if(((ItemInfo *)i1)->ol_s_w_id == ((ItemInfo *)i2)->ol_s_w_id &&
         ((ItemInfo *)i1)->ol_i_id == ((ItemInfo *)i2)->ol_i_id) {
		ret = 0;
	}else{
		ret = 1;
	}

	return ret;
}


#define NEW_ORDER_PROC "CALL NEWORDER(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
/*	
 	par_w_id BINARY_INTEGER,
	par_d_id BINARY_INTEGER,
    par_c_id BINARY_INTEGER, 
    par_o_all_local BINARY_INTEGER, 
    par_o_ol_cnt BINARY_INTEGER, 
    par_w_tax OUT BINARY_INTEGER, 
    par_d_tax OUT BINARY_INTEGER, 
    par_o_id OUT BINARY_INTEGER, 
    par_c_discount OUT BINARY_INTEGER,
    par_c_credit OUT varchar2, 
    par_c_last IN OUT varchar2,
    par_retry IN OUT BINARY_INTEGER, 
    par_cr_date DATE,
    par_ol_i_id intarray, 
    par_ol_supply_w_id intarray, 
    par_i_price IN OUT numarray, 
    par_i_name IN OUT distarray, 
    par_s_quantity IN OUT intarray, 
    par_brand_generic IN OUT chararray,
    par_ol_amount IN OUT intarray, 
    par_s_remote intarray, 
    par_ol_quantity intarray
*/

#define PAYMENT_PROC "CALL PAYMENT(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
/*
    ware_id INTEGER, 
    dist_id INTEGER,
    cust_w_id INTEGER,
    cust_d_id INTEGER,
    cust_id IN OUT INTEGER,
    bylastname INTEGER,
    hist_amount NUMBER,
    cust_last IN OUT VARCHAR2,
    ware_street_1 OUT VARCHAR2,
    ware_street_2 OUT VARCHAR2,
    ware_city OUT VARCHAR2,
    ware_state OUT VARCHAR2,
    ware_zip OUT VARCHAR2,
    dist_street_1 OUT VARCHAR2,
    dist_street_2 OUT VARCHAR2,
    dist_city OUT VARCHAR2,
    dist_state OUT VARCHAR2,
    dist_zip OUT VARCHAR2,
    cust_first OUT VARCHAR2,
    cust_middle OUT VARCHAR2,
    cust_street_1 OUT VARCHAR2,
    cust_street_2 OUT VARCHAR2,
    cust_city OUT VARCHAR2,
    cust_state OUT VARCHAR2,
    cust_zip OUT VARCHAR2,
    cust_phone OUT VARCHAR2,
    cust_since OUT DATE,
    cust_credit IN OUT VARCHAR2,
    cust_credit_lim OUT NUMBER,
    cust_discount OUT NUMBER,
    cust_balance IN OUT NUMBER,
    cust_data OUT VARCHAR2,
    cr_date IN DATE,
    retry IN OUT INTEGER
 */

#define ORDER_STATUS_PROC "CALL ORDERSTATUS(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
/*
 	ware_id INTEGER, 
    dist_id INTEGER,
    cust_id IN OUT INTEGER,
    bylastname INTEGER,
    cust_last IN OUT VARCHAR2,
    cust_first OUT VARCHAR2,
    cust_middle OUT VARCHAR2,
    cust_balance OUT NUMBER,
    ord_id IN OUT NUMBER,
    ord_entry_d OUT VARCHAR2, 
    ord_carrier_id OUT INTEGER,
    ord_ol_cnt OUT INTEGER,
    oline_supply_w_id IN OUT intarray,
    oline_i_id IN OUT intarray,
    oline_quantity IN OUT intarray,
    oline_amount IN OUT numarray,
    oline_delivery_d OUT datarray
 */

#define STOCK_LEVEL_PROC "CALL STOCKLEVEL(?,?,?,?);"
/*
    ware_id INTEGER,
    dist_id INTEGER,
    threshold INTEGER,
    low_stock OUT INTEGER
 */


bool txn_execution_trace(oceanbase_connection * obc)
{
	bool ret = true;

	if(obc->trace_info.rte_send_ts != 0 && 
		obc->trace_info.get_rte_ts != -1 &&
		obc->trace_info.get_db_ts != -1){
		int in_time = (int)(obc->trace_info.get_rte_ts - obc->trace_info.rte_send_ts);
		int db_time = (int)(obc->trace_info.get_db_ts - obc->trace_info.get_rte_ts);

		if(in_time <= 0){
			in_time = 0;
			ret = false;
		}else if(db_time <= 0){
			db_time = 0;
			ret = false;
		}else{}

		all_tpcc_txns[obc->txn_type].txn_cnt ++;
		all_tpcc_txns[obc->txn_type].total_in_time += in_time;
		if(in_time > all_tpcc_txns[obc->txn_type].max_in_time){
			all_tpcc_txns[obc->txn_type].max_in_time = in_time;
		}else{}

		all_tpcc_txns[obc->txn_type].total_db_time += db_time;
		if(db_time > all_tpcc_txns[obc->txn_type].max_db_time){
			all_tpcc_txns[obc->txn_type].max_db_time = db_time;
		}else{}
	}

	return ret;
}

/*
 *
 */

int oceanbase_get_result(oceanbase_connection * obc, ngx_log_t* log)
{
	SQLHANDLE stmt = obc->stmts[obc->txn_type];
	SQLRETURN retcode = SQL_SUCCESS;
	int ret = OCEANBASE_RETURN_OK;

	// first time to SQLExecuteAsync, record time.
	if(obc->trace_info.rte_send_ts != 0 && obc->trace_info.get_rte_ts == 0){
		struct timeval get_rte_tv;
        if(gettimeofday(&get_rte_tv, 0) < 0){
			obc->trace_info.get_rte_ts = -1;
			ngx_log_error(NGX_LOG_ERR, log, 0, "oceanbase_trace: gettimeofday failed\n");
		}else{
			obc->trace_info.get_rte_ts = TV_2_MSEC(get_rte_tv);
		}
	}else{}

	if((retcode = SQLExecuteAsync(stmt, &obc->estate)) 
						== SQL_STILL_EXECUTING){
			ret = OCEANBASE_RETURN_IO_WAIT;
	}else{
		// record ts after getting db response
		if(obc->trace_info.rte_send_ts != 0){
			struct timeval get_db_tv;
			if(gettimeofday(&get_db_tv, 0) < 0){
				obc->trace_info.get_db_ts = -1;
				ngx_log_error(NGX_LOG_ERR, log, 0, "oceanbase_trace: gettimeofday failed\n");
			}else{
				obc->trace_info.get_db_ts = TV_2_MSEC(get_db_tv);
				if(!txn_execution_trace(obc)){
					ngx_log_error(NGX_LOG_WARN, log, 0, "oceanbase_trace: unexpected elapsed time\n");
				}else{}
			}
		}else{}

		if(SQL_FAIL(retcode)){ 
			char msg[300] = {'\0'};
			
			get_sql_diag(SQL_HANDLE_STMT, stmt, msg, &obc->need_reconn);
			sprintf(obc->msg, "%s%s %s",
											TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, msg);
			ngx_log_error(NGX_LOG_ERR, log, 0, "%s%s %s\n",
											TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, msg);

			if(!obc->need_reconn){
				SQLFreeStmt(stmt, SQL_CLOSE);
				if(!rollback_on_error(obc->dbc, msg, &obc->need_reconn)){
					ngx_log_error(NGX_LOG_ERR, log, 0, "SQLEndTran failed: %s\n", msg);
				}else{}
			}else{}

			ret = OCEANBASE_RETURN_ERROR_CODE; 
		}else if(obc->txn_type == NEW_ORDER){
			int ol_cnt = obc->no_info.o_ol_cnt_b;

			if(obc->no_info.o_ol_cnt < ol_cnt){
				if(obc->no_info.o_ol_cnt != ol_cnt - 1 || 
						strcmp(obc->no_info.ol_i_name[ol_cnt - 1], "NO ITEM")){
					ngx_log_error(NGX_LOG_ERR, log, 0, "New Order failed, ol_cnt mismatch:%d/%d/%s\n",
									obc->no_info.o_ol_cnt, ol_cnt, obc->no_info.ol_i_name[ol_cnt - 1]);
					sprintf(obc->msg, "%s%s ol_cnt mismatch: a:%d/b:%d/i_name:%s w_id: %d d_id: %d c_id: %d oid: %d",
									TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, 
									obc->no_info.o_ol_cnt, ol_cnt, obc->no_info.ol_i_name[ol_cnt - 1],
									obc->no_info.w_id, obc->no_info.d_id, obc->no_info.c_id, obc->no_info.o_id);

					ret = OCEANBASE_RETURN_ERROR_CODE; 
				}else{
					sprintf(obc->msg, "%s%s", TXN_EXEC_STATE_PREFIX, INVALID_I_ID_MSG);
					ret = OCEANBASE_RETURN_OK;
				}
			}else{
				sprintf(obc->msg, "%s%s", TXN_EXEC_STATE_PREFIX, TXN_EXEC_SUCCESS_MSG);
				ret = OCEANBASE_RETURN_OK;
			}
		}else{
			sprintf(obc->msg, "%s%s", TXN_EXEC_STATE_PREFIX, TXN_EXEC_SUCCESS_MSG);
			ret = OCEANBASE_RETURN_OK;
		}

		if(obc->t_id && !check_set_terminal_state(obc->t_id, 0, log)){
			assert(false);
		}
		
		obc->total_txn_cnt ++;
		
		// reconnect db if connection is used too many times
		if(g_max_txn_cnt && obc->total_txn_cnt >= g_max_txn_cnt){
			int ret_conn = 0;

			ob_free_conn(obc);
			ret_conn = ob_connect(obc);	
			ngx_log_error(NGX_LOG_ERR, log, 0, "Reconnect max_txn_cnt: %d conn_ret: %d\n", g_max_txn_cnt, ret_conn);
		}else{}
	}

	return ret;
}

int new_order(ItemInfo * items, oceanbase_connection * obc, ngx_log_t* log)
{
	int ret = OCEANBASE_RETURN_OK;
	int i = 0;

	assert(items != NULL);
	qsort(items, obc->no_info.o_ol_cnt, sizeof(ItemInfo), item_cmp);
	obc->no_info.alocal = true;

	for(i = 0; i < obc->no_info.o_ol_cnt; i ++){
		obc->no_info.ol_s_w_id[i] = items[i].ol_s_w_id;
		obc->no_info.ol_i_id[i] = items[i].ol_i_id;
		obc->no_info.ol_quantity[i] = items[i].ol_quantity;
		if(obc->no_info.ol_s_w_id[i] != obc->no_info.w_id){
			obc->no_info.alocal = false;
			obc->no_info.sremote[i] = 1;
		}else{
			obc->no_info.sremote[i] = 0;
		}
	}
	
	ngx_log_error(NGX_LOG_DEBUG, log, 0, "new_order before SQLExecute cnt: %d i_id[0]:%d\n", 
												obc->no_info.o_ol_cnt, obc->no_info.ol_i_id[0]);
	
	obc->no_param.ol_i_id_arr.size = obc->no_info.o_ol_cnt;
	obc->no_param.ol_s_w_id_arr.size = obc->no_info.o_ol_cnt;
	obc->no_param.ol_qty_arr.size = obc->no_info.o_ol_cnt;
	obc->no_param.sremote_arr.size = obc->no_info.o_ol_cnt;

	ret = oceanbase_get_result(obc, log);
    	
	return ret;
}

int process_new_order(oceanbase_connection * obc, ngx_log_t* log)
{
	int ret = 0;
	int t_id = 0;
	int w_id = 0;
	int	sl_d_id = 0;
	int d_id = 0;
	int c_id = 0;
	
	int ol_cnt_b = 0;
	int o_len = 0;
	bool	bnull = false;
	char err_msg[128] = {'\0'};
	ItemInfo items[MAX_OL_NUM];
	NewOrderInfo* no_info = &(obc->no_info);
	memset(no_info, 0, sizeof(NewOrderInfo));
	u_char *params = (u_char *)(obc->params);

	char *s_w[15] = {"&s00=", "&s01=", "&s02=", "&s03=", "&s04=", "&s05=", "&s06=", 
					 "&s07=", "&s08=", "&s09=", "&s10=", "&s11=", "&s12=", "&s13=", "&s14="};
	char *s_i[15] = {"&i00=", "&i01=", "&i02=", "&i03=", "&i04=", "&i05=", "&i06=", 
					 "&i07=", "&i08=", "&i09=", "&i10=", "&i11=", "&i12=", "&i13=", "&i14="};
	char *s_q[15] = {"&q00=", "&q01=", "&q02=", "&q03=", "&q04=", "&q05=", "&q06=", 
					 "&q07=", "&q08=", "&q09=", "&q10=", "&q11=", "&q12=", "&q13=", "&q14="};

	if(params == NULL){
		sprintf(err_msg, "need logon before executing transactions");
		ret = -1;
	}else{
		if(ngx_strncmp(params, "t_id=", 5) !=0 ){
			sprintf(err_msg, "need logon before executing transactions");
			ret = -1;
		}else{
			params += 5;
			parse_param(params, t_id, bnull);
			obc->t_id = no_info->t_id = t_id;
			if(bnull || ngx_strncmp(params, "&sl_d_id=", 9) != 0){
				sprintf(err_msg, "need logon before executing transactions");
				ret =  -1;
			}else{
				params += 9;
				parse_param(params, sl_d_id, bnull);
				no_info->sl_d_id = sl_d_id;
				if(bnull || ngx_strncmp(params, "&w_id=", 6) != 0){
					sprintf(err_msg, "need logon before executing transactions");
					ret = -1;
				}else{
					params += 6;
					parse_param(params, w_id, bnull);	
					no_info->w_id = w_id;
					if(bnull || ngx_strncmp(params, "&d_id=", 6) != 0){
						sprintf(err_msg, "need logon before executing transactions");
						ret = -1;
					}else{
						params += 6;
						parse_param(params, d_id, bnull);
						if(bnull || ngx_strncmp(params, "&c_id=", 6) != 0){
							sprintf(err_msg, "d_id is invalid, should be positive integer");
							ret = -1;
						}else{
							int i = 0, n = 0;
							params += 6;
							parse_param(params, c_id, bnull);
							if(bnull || c_id <= 0){
								sprintf(err_msg, "c_id is invalid, should be positive integer");
								ret = -1;
							}else{}

							for(i = 0; *params != '\0' && !ret; i++){
								bool bnull_arr[3] = {false, false, false};
								
								if(ngx_strncmp(params, "&trace=", 7) == 0){
									long long int ts = 0;
									bool tmp_bnull = false;
									params += 7;
									parse_param(params, ts, tmp_bnull);
									obc->trace_info.rte_send_ts = ts;
									assert(!tmp_bnull);
									break;
								}

								if(ngx_strncmp(params, s_w[i], 5) != 0){
									ret = -1;
								}else{
									params += 5;
									parse_param(params, items[n].ol_s_w_id, bnull_arr[0]);
									if(ngx_strncmp(params, s_i[i], 5) != 0){
										ret = -1;
									}else{
										params += 5;
										parse_param(params, items[n].ol_i_id, bnull_arr[1]);
										if(ngx_strncmp(params, s_q[i], 5) != 0){
											ret = -1;
										}else{
											params += 5;
											parse_param(params, items[n].ol_quantity, bnull_arr[2]);
											ngx_log_error(NGX_LOG_DEBUG, log, 0, "idx: %d swid: %d iid:%d qty: %d\n",
																i, items[n].ol_s_w_id, items[n].ol_i_id, items[n].ol_quantity);
											if(bnull_arr[0] && bnull_arr[1] && bnull_arr[2]){
											}else if(!bnull_arr[0] && !bnull_arr[1] && !bnull_arr[2]){
												if((items[n].ol_s_w_id <= 0) 
																	|| (items[n].ol_i_id <= 0) 
																	|| (items[n].ol_quantity <= 0)){ 
													sprintf(err_msg, "invalid item information: %d %d %d, "
																	 "should be positive integer", 
																	items[n].ol_s_w_id,
																	items[n].ol_i_id,
																	items[n].ol_quantity);
													ret = -1;
												}else{
													n ++;
												}
											}else{
												ret = -1;
											}
										}
									}
								}
							} // end for
							
							i = n;
							
							if (i == 0 && ret == 0){
								sprintf(err_msg, "no item assigned to execute new-order transaction");
								ret = -1;
							}else if(d_id < 1){
								sprintf(err_msg, "d_id is invalid, should be positive integer");
								ret = -1;
							}else{}
						
							ngx_log_error(NGX_LOG_DEBUG, log, 0, "new_order: ret:%d item_cnt: %d sizeof(s_i[0]):%d \n", 
																ret, i, sizeof(s_i[0]));	
							// params ok, go on
							if( ret == 0 ){
								if( !check_set_terminal_state(t_id, 1, log) ){
									obc->t_id = 0;
									ret = -2;
									ngx_log_error(NGX_LOG_ERR, log, 0, "new_order: check_set_terminal_state failed\n");
									o_len += sprintf(obc->msg, "%s%s new_order transaction terminal %d is busy", 
																	TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, t_id);
								}else{
									no_info->d_id = d_id;
									no_info->c_id = c_id;
									no_info->o_ol_cnt = i;
									no_info->o_ol_cnt_b = i;
									
									ol_cnt_b = i;

									ngx_log_error(NGX_LOG_DEBUG, log, 0, "new_order:t_id=%d&w_id=%d&d_id=%d&c_id=%d"
														"&ol_cnt=%d&s_w_id[0]=%d&s_i_id[0]=%d&s_i_q[0]=%d\n", 
															t_id, w_id, d_id, c_id, ol_cnt_b, items[0].ol_s_w_id, 
															items[0].ol_i_id, items[0].ol_quantity);
									
									if(!get_timestamp(no_info->o_entry_d)){
										ret = -3;
										o_len += sprintf(obc->msg, "%s%s get_timestamp failed",
																		TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG);
									}else if((ret = new_order(items, obc, log)) == OCEANBASE_RETURN_OK){
										ngx_log_error(NGX_LOG_WARN, log, 0, "Immediate|new_order:w_id=%d and d_id=%d and c_id=%d"
														" and o_id = %d and ol_cnt=%d c_credit: %s\n", 
															w_id, d_id, c_id, no_info->o_id, no_info->o_ol_cnt, 
															no_info->c_credit); 

									}else{}
								}
							}else{}
						}
					}
				}
			}
		}
	}

	if(ret == -1){
		o_len += sprintf(obc->msg, "%s%s new_order transaction parameters: %s <br>",
										TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, err_msg);
	}else{}

	if(ret < 0){
		ret = OCEANBASE_RETURN_ERROR_CODE;
	}

	return ret;
}

#define NEW_ORDER_OUTPUT_BUF_SIZE  6144
#define PAYMENT_OUTPUT_BUF_SIZE  4096 
#define DELIVERY_OUTPUT_BUF_SIZE 2048 
#define ORDER_STATUS_OUTPUT_BUF_SIZE  6144
#define STOCK_LEVEL_OUTPUT_BUF_SIZE  2048 

static char *txn_menu =
				"<pre>"
                "<h3>TPC-C Transaction Menu</h3>"
                "<a href=\"/tpcc/new-order-input?t_id=%d&sl_d_id=%d&w_id=%d\">new-order</a><br>"
                "<a href=\"/tpcc/payment-input?t_id=%d&sl_d_id=%d&w_id=%d\">payment</a><br>"
                "<a href=\"/tpcc/delivery-input?t_id=%d&sl_d_id=%d&w_id=%d\">delivery</a><br>"
                "<a href=\"/tpcc/order-status-input?t_id=%d&sl_d_id=%d&w_id=%d\">order-status</a><br>"
                "<a href=\"/tpcc/stock-level-input?t_id=%d&sl_d_id=%d&w_id=%d\">stock-level</a><br>"
				"<a href=\"/tpcc/logon\">exit</a><br>"
				"</pre>";

extern int trans_html_string(char* dst, char* src, int ilen);

ngx_int_t
ngx_http_oceanbase_output_new_order(ngx_http_request_t *r, bool succ)
{
    u_char  *pos;
    char *out;
    ngx_http_upstream_t  *u = r->upstream;
    size_t  size;
    ngx_http_upstream_oceanbase_peer_data_t   *obp = u->peer.data;
    oceanbase_connection                          *obc = obp->oceanbase_con;
    int o_len = 0;
    NewOrderInfo * no_info = &(obc->no_info);
    ngx_log_t*  log = r->connection->log;
    size = NEW_ORDER_OUTPUT_BUF_SIZE;

    pos = ngx_http_oceanbase_request_mem(r, obp, size);
    if (pos == NULL) {
        return NGX_ERROR;
    }

    ngx_memzero(pos, size);

    out = (char *)pos;
    o_len += gen_response_head(out, "New Order");

    if(succ){
        o_len += sprintf(out + o_len, "Warehouse: %10d  District: %2d Date: %s <br>",
                                        no_info->w_id, no_info->d_id, no_info->o_entry_d);
        o_len += sprintf(out + o_len, "Customer: %d Name: ", no_info->c_id);
        o_len += trans_html_string(out + o_len, no_info->c_last, 18);
        o_len += sprintf(out + o_len, "Credit: %s Discount: %05.4f <br>",
                                                no_info->c_credit, no_info->c_discount);
        o_len += sprintf(out + o_len, "Order Number: %d Number of Lines: %d W_tax: %.4f D_tax: %.4f <br>",
                            no_info->o_id, no_info->o_ol_cnt, no_info->w_tax, no_info->d_tax);
        o_len += sprintf(out + o_len, "Supp-W     Item-Id    Item-Name                 Qty  Stock     ");
        o_len += sprintf(out + o_len, "B/G   Price    Amount<br>");

        if( no_info->o_ol_cnt_b == no_info->o_ol_cnt ){
            double total_amount = 0.0;
            int j = 0;

            for(j = 0; j < no_info->o_ol_cnt; j ++){
                o_len += sprintf(out + o_len, "%-10d %-6d     ", no_info->ol_s_w_id[j],
                                                no_info->ol_i_id[j]);
                o_len += trans_html_string(out + o_len, no_info->ol_i_name[j], 26);
                o_len += sprintf(out + o_len, "%-2d   %-10d  %s   %-6.2f   %-10.2f  <br>",
                no_info->ol_quantity[j], no_info->ol_stock[j], no_info->ol_b_g[j],
                no_info->ol_i_price[j], no_info->ol_amount[j]);
                total_amount += no_info->ol_amount[j];
            }

            total_amount *= (1 - no_info->c_discount) * (1 + no_info->w_tax) * (1 + no_info->d_tax);
            o_len += sprintf(out + o_len, "Total: %.2f<br>", total_amount);
            //o_len += sprintf(out + o_len, "Total: %.2f<br>", total_amount);
        }else{}
    }else{
        //ngx_log_error(NGX_LOG_ERR, log, 0, "%s\n", no_info->status);
    }

    o_len += sprintf(out + o_len, "%s <br>", obc->msg);
    o_len += sprintf(out + o_len, txn_menu, no_info->t_id, no_info->sl_d_id, no_info->w_id, 
											no_info->t_id, no_info->sl_d_id, no_info->w_id,
											no_info->t_id, no_info->sl_d_id, no_info->w_id,
											no_info->t_id, no_info->sl_d_id, no_info->w_id,
											no_info->t_id, no_info->sl_d_id, no_info->w_id);
    o_len += sprintf(out + o_len, "%s", response_footer);

	if(o_len >= (int)(size - 1)){
		ngx_log_error(NGX_LOG_ERR, log, 0, "new-order output len: %d buffer size: %d\n", o_len, size);
		assert(false);
	}
    obp->seen_stream_end = 1;

    return ngx_http_oceanbase_submit_mem(r, obp, o_len + 1);
}

int process_payment(oceanbase_connection * obc, ngx_log_t* log)
{
	int ret = 0;
	int t_id = 0;
	int w_id = 0;
	int	sl_d_id = 0;
	int d_id = 0;
	int c_id = 0;
	u_char c_last[20] = {0};
	int c_w_id = 0;
	int c_d_id = 0;
	double h_amount = 0.0;
	int o_len = 0;
	bool bnull = false, bnull_cid = false;
	char err_msg[128] = {'\0'};
	u_char *params = (u_char *)(obc->params);
	PaymentInfo *pay_info = &(obc->pay_info);
	memset(pay_info, 0, sizeof(PaymentInfo));

    if(params == NULL){
		sprintf(err_msg, "need logon before executing transactions");
        ret = -1;
    }else{
        if(ngx_strncmp(params, "t_id=", 5) !=0 ){
			sprintf(err_msg, "need logon before executing transactions");
            ret = -1;
        }else{
			params += 5;
            parse_param(params, t_id, bnull);
			obc->t_id = pay_info->t_id = t_id;
            if(bnull || ngx_strncmp(params, "&sl_d_id=", 9) != 0){
				sprintf(err_msg, "need logon before executing transactions");
                ret =  -1;
            }else{
				params += 9;
                parse_param(params, sl_d_id, bnull);
				pay_info->sl_d_id = sl_d_id;
                if(bnull || ngx_strncmp(params, "&w_id=", 6) != 0){
					sprintf(err_msg, "need logon before executing transactions");
                    ret = -1;
                }else{
					params += 6;
                    parse_param(params, w_id, bnull);
					pay_info->w_id = w_id;
					if(bnull || ngx_strncmp(params, "&d_id=", 6) != 0){
						sprintf(err_msg, "need logon before executing transactions");
						ret = -1;
					}else{
						params += 6;
						parse_param(params, d_id, bnull);
						if(bnull || ngx_strncmp(params, "&c_id=", 6) != 0){
							sprintf(err_msg, "d_id is invalid, should be positive integer");
							ret = -1;
						}else{
							params += 6;
							parse_param(params, c_id, bnull_cid);
							if(ngx_strncmp(params, "&c_w_id=", 8) != 0){
								ret = -1;
							}else{
								params += 8;
								parse_param(params, c_w_id, bnull);
								if(bnull || ngx_strncmp(params, "&c_d_id=", 8) != 0){
									sprintf(err_msg, "c_w_id is invalid, should be positive integer");
									ret = -1;
								}else{
									params += 8;
									parse_param(params, c_d_id, bnull);
									if(bnull || ngx_strncmp(params, "&c_last=", 8) != 0){
										sprintf(err_msg, "c_d_id is invalid, should be positive integer");
										ret = -1;
									}else{
										int i = 0;
										params += 8;
										for(i =0; *params != '\0' && *params != '&'; i ++,params ++){
											;
										}
										if(i > 0){
											ngx_cpystrn(c_last, params - i, i + 1); //unexcepted ngx_cpystrn len
										}else{}

										if(!i && bnull_cid){
											sprintf(err_msg, "one of c_id and c_last_name should be assigned");
											ret = -1;
										}else if (!bnull_cid && c_id < 1){
											sprintf(err_msg, "c_id is invalid, should be positive integer");
											ret = -1;
										}else{}
										
										if(ret == -1 || (ngx_strncmp(params, "&h_amount=", 10) != 0)){
											ret = -1;
										}else{
											params += 10;
											parse_param(params, h_amount, bnull);

											if(ngx_strncmp(params, "&trace=", 7) == 0){
												long long int ts = 0;
												bool tmp_bnull = false;
												params += 7;
												parse_param(params, ts, tmp_bnull);
												obc->trace_info.rte_send_ts = ts;
												assert(!tmp_bnull);
											}else{}

											if(bnull || h_amount < 0.00 ){
												sprintf(err_msg, "h_amount is invalid, should be positive");
												ret = -1;
											}else if(*params != '\0'){
												sprintf(err_msg, "invalid parameters");
												ret = -1;
											}else if( d_id <= 0 ){
												sprintf(err_msg, "d_id is invalid, should be positive integer");
												ret = -1;
											}else if( c_w_id <= 0 ){
												sprintf(err_msg, "c_w_id is invalid, should be positive integer");
												ret = -1;
											}else if (c_d_id <= 0 ){
												sprintf(err_msg, "c_d_id is invalid, should be positive integer");
												ret = -1;
											}else if( check_set_terminal_state(t_id, 1, log) ){
												pay_info->d_id = d_id;
												pay_info->c_id = c_id;
												pay_info->c_w_id = c_w_id;
												pay_info->c_d_id = c_d_id;
												pay_info->h_amount = h_amount;
												if(bnull_cid){
													strcpy(pay_info->c_last, (char *)c_last);
												}else{}
												pay_info->blastname = bnull_cid;

												ngx_log_error(NGX_LOG_DEBUG, log, 0, "process_payment:t_id=%d&w_id=%d&d_id=%d"
													"&c_id=%d&c_last=%s&c_w_id=%d&c_d_id=%d&h_amount=%.2f\n", 
															t_id, w_id, d_id, c_id, c_last, c_w_id, c_d_id, h_amount);
												if(!get_timestamp(pay_info->h_date)){
													ret = -2;
													o_len += sprintf(obc->msg + o_len, "get_timestamp failed");	
												}else{ 
													if((ret = oceanbase_get_result(obc, log)) == OCEANBASE_RETURN_OK){
														ngx_log_error(NGX_LOG_WARN, log, 0, 
																		"Immediate|process_payment:w_id=%d&d_id=%d"
																		"&c_id=%d&c_last=%s&c_w_id=%d&c_d_id=%d&h_amount=%.2f\n", 
																		w_id, d_id, c_id, c_last, c_w_id, c_d_id, h_amount);
													}
												}
											}else{
												ret = -2;
												obc->t_id = 0;
												o_len += sprintf(obc->msg + o_len, 
															"%s%s payment transaction terminal %d is busy", 
															TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, 
																			t_id);
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
	if(ret == -1){
		o_len += sprintf(obc->msg + o_len, "%s%s payment transaction parameters error: %s", 
										TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, err_msg);
	}else{}

    if(ret < 0){
        ret = OCEANBASE_RETURN_ERROR_CODE;
    }

	return ret;
}

ngx_int_t
ngx_http_oceanbase_output_payment(ngx_http_request_t *r, bool succ)
{
    u_char  *pos;
    char *out;
    ngx_http_upstream_t  *u = r->upstream;
    size_t  size;
    ngx_http_upstream_oceanbase_peer_data_t   *obp = u->peer.data;
    oceanbase_connection                          *obc = obp->oceanbase_con;
	ngx_log_t*  log = r->connection->log;
    int o_len = 0;
    PaymentInfo * pay_info = &(obc->pay_info);
    size = PAYMENT_OUTPUT_BUF_SIZE;

    pos = ngx_http_oceanbase_request_mem(r, obp, size);
    if (pos == NULL) {
        return NGX_ERROR;
    }

    ngx_memzero(pos, size);

    out = (char *)pos;
    o_len += gen_response_head(out, "Payment");

	if(succ){
		char temp1[40] ={0};
		char temp2[40] ={0};
		char temp3[40] ={0};
		char temp4[40] ={0};
		o_len += sprintf(out + o_len, "Date: %s <br>", pay_info->h_date); 
		o_len += sprintf(out + o_len, "Warehouse: %-10d District: %-2d <br>", 
											pay_info->w_id, pay_info->d_id); 
		o_len += trans_html_string(out + o_len, pay_info->w_street_1, 22);
		o_len += trans_html_string(out + o_len, pay_info->d_street_1, 22);
		o_len += sprintf(out + o_len, "<br>");
		o_len += trans_html_string(out + o_len, pay_info->w_street_2, 22);
		o_len += trans_html_string(out + o_len, pay_info->d_street_2, 22);
		o_len += sprintf(out + o_len, "<br>");
		o_len += trans_html_string(out + o_len, pay_info->w_city, 22);
		o_len += trans_html_string(out + o_len, pay_info->d_city, 22);
		o_len += sprintf(out + o_len, "<br>");
		o_len += trans_html_string(out + o_len, pay_info->w_state, 22);
		o_len += trans_html_string(out + o_len, pay_info->d_state, 22);
		o_len += sprintf(out + o_len, "<br>");
		o_len += trans_html_string(out + o_len, pay_info->w_zip, 22);
		o_len += trans_html_string(out + o_len, pay_info->d_zip, 22);
		o_len += sprintf(out + o_len, "<br>");
		o_len += sprintf(out + o_len, "Customer: %-4d  Cust-Warehouse: %-10d Cust-District: %-2d<br>"
							, pay_info->c_id, pay_info->c_w_id, pay_info->c_d_id);
		o_len += sprintf(out + o_len, "Name: ");
		o_len += trans_html_string(out + o_len, pay_info->c_first, 18);
		o_len += trans_html_string(out + o_len, pay_info->c_middle, 4);
		o_len += trans_html_string(out + o_len, pay_info->c_last, 18);
		o_len += sprintf(out + o_len, "Since: ");
		o_len += trans_html_string(out + o_len, pay_info->c_since, 12);
		
		trans_html_string(temp1, pay_info->c_street_1, 18);
		o_len += sprintf(out + o_len, "<br>%s Credit: %s", temp1, pay_info->c_credit);
		trans_html_string(temp1, pay_info->c_street_2, 18);
		o_len += sprintf(out + o_len, "<br>%s %%Discount: %.4f<br>", temp1, pay_info->c_discount);
		trans_html_string(temp1, pay_info->c_city, 22);
		trans_html_string(temp2, pay_info->c_state, 4);
		trans_html_string(temp3, pay_info->c_zip, 12);
		trans_html_string(temp4, pay_info->c_phone, 18);
		o_len += sprintf(out + o_len, "%s %s %s %s<br>"
									 "Amount Paid: $%.2f  Balance: $%.2f<br>"
									"Credit Limit: $%.2f <br>", temp1, temp2, temp3, temp4,
								pay_info->h_amount, pay_info->c_balance, pay_info->c_credit_lim);
		o_len += sprintf(out + o_len, "Cust-Data: ");
		o_len += trans_html_string(out + o_len, pay_info->c_data, 201);
		o_len += sprintf(out + o_len, "<br>");
	}else{
        //ngx_log_error(NGX_LOG_ERR, log, 0, "%s\n", pay_info->status);
    }

	o_len += sprintf(out + o_len, "%s <br>", obc->msg);
    o_len += sprintf(out + o_len, txn_menu, pay_info->t_id, pay_info->sl_d_id, pay_info->w_id,
                                            pay_info->t_id, pay_info->sl_d_id, pay_info->w_id,
                                            pay_info->t_id, pay_info->sl_d_id, pay_info->w_id,
                                            pay_info->t_id, pay_info->sl_d_id, pay_info->w_id,
                                            pay_info->t_id, pay_info->sl_d_id, pay_info->w_id);
    o_len += sprintf(out + o_len, "%s", response_footer);

    if(o_len >= (int)(size -1)){
        ngx_log_error(NGX_LOG_ERR, log, 0, "payment output len: %d buffer size: %d\n", o_len, size);
        assert(false);
    }

    obp->seen_stream_end = 1;

    return ngx_http_oceanbase_submit_mem(r, obp, o_len + 1);
}

//#define DELIVERY_PROC "CALL DELIVERY(?,?,?,?,?,?,?,?,?);"
/*
	ware_id IN INTEGER,
    dist_id IN OUT intarray,
    order_id OUT intarray,
    ordcnt OUT INTEGER,
    sums OUT intarray,
    del_date IN DATE,
    carrier_id IN INTEGER,
    order_c_id OUT intarray, 
    retry IN OUT INTEGER, 
 */

int process_delivery(u_char *params, char *out, int len, ngx_log_t *log)
{
    int t_id = 0;
    int w_id = 0;
	int sl_d_id = 0;
    int o_carrier_id = 0;
    int ret = 0;
	char err_msg[128] = {'\0'};
	bool bnull = false;
	long long int ts = 0;
	int	dtt = 0;

    int o_len = 0;
    o_len = gen_response_head(out, "Delivery");

    if(params == NULL || out == NULL || len <= 0){
		sprintf(err_msg, "need logon before executing transactions");
        ret = -1;
    }else{
        if(ngx_strncmp(params, "t_id=", 5) !=0 ){
			sprintf(err_msg, "need logon before executing transactions");
            ret = -1;
        }else{
            params += 5;
            parse_param(params, t_id, bnull);
            if(bnull || ngx_strncmp(params, "&sl_d_id=", 9) != 0){
				sprintf(err_msg, "need logon before executing transactions");
                ret =  -1;
            }else{
                params += 9;
                parse_param(params, sl_d_id, bnull);
				if(bnull || ngx_strncmp(params, "&w_id=", 6) != 0){
					sprintf(err_msg, "need logon before executing transactions");
					ret =  -1;
				}else{
					params += 6;
					parse_param(params, w_id, bnull);
					if(bnull || ngx_strncmp(params, "&o_carrier_id=", 14) != 0){
						sprintf(err_msg, "need logon before executing transactions");
						ret = -1;
					}else{
						params += 14;
						parse_param(params, o_carrier_id, bnull);

						if(ngx_strncmp(params, "&dtt=", 5) == 0){
							bool tmp_bnull = false;
							params += 5;
							parse_param(params, dtt, tmp_bnull);
							assert(tmp_bnull || !tmp_bnull); //silent compiler
						}else{}

						if(ngx_strncmp(params, "&trace=", 7) == 0){
							bool tmp_bnull = false;
							params += 7;
							parse_param(params, ts, tmp_bnull);
							assert(!tmp_bnull);
						}else{}

						if(bnull || *params != '\0' || o_carrier_id <= 0){
							sprintf(err_msg, "carrier_id is invalid, should be positive integer");
							ret = -1;
						}else if( check_set_terminal_state(t_id, 1, log) ){
							char status[30];
							ngx_log_error(NGX_LOG_DEBUG, log, 0, "delivery:t_id=%d&w_id=%d&o_carrier_id=%d\n",
																					t_id, w_id, o_carrier_id);
							if( !put_delivery_task(w_id, o_carrier_id, dtt, ts, (char*)status, log) ){
								ret = -2;
								o_len += sprintf(out + o_len, "%s%s %s <br>", TXN_EXEC_STATE_PREFIX,
																		TXN_EXEC_FAILURE_MSG, status);
							}else{
								o_len += sprintf(out + o_len, "Warehouse: %d <br>", w_id);
								o_len += sprintf(out + o_len, "Carrier Number: %d <br>", o_carrier_id);
								o_len += sprintf(out + o_len, "%s%s <br>", TXN_EXEC_STATE_PREFIX, status);
							}

							if( !check_set_terminal_state(t_id, 0, log) ){
								assert(false);
							}
						}else{
							o_len += sprintf(out + o_len, "%s%s delivery transaction terminal %d is busy",
															TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, t_id);
							ret = -3;
						}
					}
				}
            }
        }
    }

    if( ret == -1){
        o_len += sprintf(out + o_len, "%s%s delivery transaction parameters error: %s",
                                            TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, err_msg);
    }else{}
    o_len += sprintf(out + o_len, txn_menu, t_id, sl_d_id, w_id,
                                            t_id, sl_d_id, w_id,
                                            t_id, sl_d_id, w_id,
                                            t_id, sl_d_id, w_id,
                                            t_id, sl_d_id, w_id);
    o_len += sprintf(out + o_len, "%s", response_footer);

    return ret;
}

int process_order_status(oceanbase_connection * obc, ngx_log_t *log)
{
    int t_id = 0;
    int w_id = 0;
    int sl_d_id = 0;
    int d_id = 0;
	int c_id = 0;
	u_char c_last[20] = {0};
	char err_msg[128] = {'\0'};
    int ret = 0;
	bool bnull = false;

	OrderStatusInfo* os_info = &(obc->os_info);
	memset(os_info, 0, sizeof(OrderStatusInfo));
    u_char *params = (u_char *)(obc->params);

	if(params == NULL){
		sprintf(err_msg, "need logon before executing transactions");
        ret = -1;
    }else{
        if(ngx_strncmp(params, "t_id=", 5) !=0 ){
			sprintf(err_msg, "need logon before executing transactions");
            ret = -1;
        }else{
            params += 5;
            parse_param(params, t_id, bnull);
			obc->t_id = os_info->t_id = t_id;
            if(bnull || ngx_strncmp(params, "&sl_d_id=", 9) != 0){
				sprintf(err_msg, "need logon before executing transactions");
                ret =  -1;
            }else{
                params += 9;
                parse_param(params, sl_d_id, bnull);
				os_info->sl_d_id = sl_d_id;
				if(bnull || ngx_strncmp(params, "&w_id=", 6) != 0){
					sprintf(err_msg, "need logon before executing transactions");
					ret = -1;
				}else{
					params += 6;
					parse_param(params, w_id, bnull);
					os_info->w_id = w_id;
					if(bnull || ngx_strncmp(params, "&d_id=", 6) != 0){
						sprintf(err_msg, "need logon before executing transactions");
						ret = -1;
					}else{
						params += 6;
						parse_param(params, d_id, bnull);
						if(bnull || d_id < 1 || ngx_strncmp(params, "&c_id=", 6) != 0){
							sprintf(err_msg, "d_id is invalid, should be positive integer");
							ret = -1;
						}else{
							params += 6;
							parse_param(params, c_id, bnull);
							if(ngx_strncmp(params, "&c_last=", 8) != 0){
								ret = -1;
							}else{
								int i =0;
								params += 8;
								for(i =0; *params != '\0' && *params != '&'; i ++,params ++){
									;
								}

								if(i > 0){
									ngx_cpystrn(c_last, params - i, i + 1);
								}else{}

								if(ngx_strncmp(params, "&trace=", 7) == 0){
									long long int ts = 0;
									bool tmp_bnull = false;
									params += 7;
									parse_param(params, ts, tmp_bnull);
									obc->trace_info.rte_send_ts = ts;
									assert(!tmp_bnull);
								}else{}

								if(*params != '\0' || (bnull && !i)){
									sprintf(err_msg, "one of c_id and c_last_name should be assigned");
									ret = -1;
								}else if(!bnull && c_id < 1 ){
									sprintf(err_msg, "c_id is invalid, should be positive integer");
									ret = -1;
								}else if( check_set_terminal_state(t_id, 1, log) ){ 
									os_info->d_id = d_id;
									os_info->c_id = c_id;
									os_info->blastname = bnull;
									if(bnull){
										strcpy(os_info->c_last, (char *)c_last);
									}else{}
									ngx_log_error(NGX_LOG_DEBUG, log, 0, 
												"order_status:t_id=%d&w_id=%d&d_id=%d&c_id=%d&c_last=%s\n",
												t_id, w_id, d_id, c_id, c_last);
									if((ret = oceanbase_get_result(obc, log)) == OCEANBASE_RETURN_OK){
										ngx_log_error(NGX_LOG_WARN, log, 0, 
												"Immediate|order_status:w_id=%d&d_id=%d&c_id=%d&c_last=%s\n",
												w_id, d_id, c_id, c_last);
									}

								}else{ 
									ret = -2;
									obc->t_id = 0;
									sprintf(obc->msg, "%s%s order_status transaction terminal %d is busy", 
												TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, t_id);
								}
							}
						}
					}
				}
			}
		}
	}

	if( ret == -1 ){
		sprintf(obc->msg, "%s%s order_status transaction parameters error: %s",
										TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, err_msg);
	}else{}
	
	if(ret < 0){
		ret = OCEANBASE_RETURN_ERROR_CODE;
	}
	
	return ret;
}

ngx_int_t
ngx_http_oceanbase_output_order_status(ngx_http_request_t *r, bool succ)
{
    u_char  *pos;
    char *out;
    ngx_http_upstream_t  *u = r->upstream;
    size_t  size;
    ngx_http_upstream_oceanbase_peer_data_t   *obp = u->peer.data;
    oceanbase_connection                          *obc = obp->oceanbase_con;
    int o_len = 0;
	ngx_log_t*  log = r->connection->log;
    OrderStatusInfo * os_info = &(obc->os_info);
    size = ORDER_STATUS_OUTPUT_BUF_SIZE;

    pos = ngx_http_oceanbase_request_mem(r, obp, size);
    if (pos == NULL) {
        return NGX_ERROR;
    }

    ngx_memzero(pos, size);

    out = (char *)pos;
    o_len += gen_response_head(out, "Order Status");

	if(succ){
		char temp1[40]={0};
		char temp2[40]={0};
		char temp3[40]={0};
		int	j = 0;

		o_len += sprintf(out + o_len, "Warehouse: %d District: %d<br>",
							os_info->w_id, os_info->d_id);
		trans_html_string(temp1, os_info->c_first, 18);
		trans_html_string(temp2, os_info->c_middle, 4);
		trans_html_string(temp3, os_info->c_last, 18);
		o_len += sprintf(out + o_len, "Customer: %d Name: %s %s %s <br>",
						 os_info->c_id, temp1, temp2, temp3);
		o_len += sprintf(out + o_len, "Balance: $%.2f <br>", os_info->c_balance);
		o_len += sprintf(out + o_len, "Order Number: %d Entry Date: %s Carrier Number: %d <br>"
									, os_info->o_id, os_info->o_entry_d, os_info->o_carrier_id);
		o_len += sprintf(out + o_len, "Supp-W     Item-Id   Qty   Amount	 Delivery-Date <br>");
		
		for(j = 0; j < os_info->o_ol_cnt; j ++){
			o_len += sprintf(out + o_len, "%-10d %-6d    %-2d	   %-10.2f    %s <br>", 
								os_info->ol_s_w_id[j],
								os_info->ol_i_id[j], 
								os_info->ol_quantity[j], 
								os_info->ol_amount[j],
								os_info->ol_delivery_d[j]);
		}
	}else{
        //ngx_log_error(NGX_LOG_ERR, log, 0, "%s\n", os_info->status);
    }

	o_len += sprintf(out + o_len, "%s <br>", obc->msg);
    o_len += sprintf(out + o_len, txn_menu, os_info->t_id, os_info->sl_d_id, os_info->w_id,
                                            os_info->t_id, os_info->sl_d_id, os_info->w_id,
                                            os_info->t_id, os_info->sl_d_id, os_info->w_id,
                                            os_info->t_id, os_info->sl_d_id, os_info->w_id,
                                            os_info->t_id, os_info->sl_d_id, os_info->w_id);
    o_len += sprintf(out + o_len, "%s", response_footer);

    if(o_len >= (int)(size - 1)){
        ngx_log_error(NGX_LOG_ERR, log, 0, "order-status output len: %d buffer size: %d\n", o_len, size);
        assert(false);
    }

    obp->seen_stream_end = 1;

    return ngx_http_oceanbase_submit_mem(r, obp, o_len + 1);
}

int process_stock_level(oceanbase_connection * obc, ngx_log_t *log)
{
    int ret = 0;
	int t_id = 0;
	char err_msg[128] = {'\0'};
	bool bnull = false;

	StockLevelInfo* sl_info = &(obc->sl_info);
    u_char *params = (u_char *)(obc->params);

    if(params == NULL){
		sprintf(err_msg, "need logon before executing transactions");
        ret = -1;
    }else{
        if(ngx_strncmp(params, "t_id=", 5) !=0 ){
			sprintf(err_msg, "need logon before executing transactions");
            ret = -1;
        }else{
            params += 5;
            parse_param(params, t_id, bnull);
			obc->t_id = sl_info->t_id = t_id;
            if(bnull || ngx_strncmp(params, "&sl_d_id=", 9) != 0){
				sprintf(err_msg, "need logon before executing transactions");
                ret =  -1;
            }else{
                params += 9;
                parse_param(params, sl_info->d_id, bnull);
                if(bnull || ngx_strncmp(params, "&w_id=", 6) != 0){
					sprintf(err_msg, "need logon before executing transactions");
                    ret = -1;
                }else{
					params += 6;
					parse_param(params, sl_info->w_id, bnull);
					if(bnull || ngx_strncmp(params, "&d_id=", 6) != 0){
						sprintf(err_msg, "need logon before executing transactions");
						ret = -1;
					}else{
						params += 6;
						parse_param(params, sl_info->d_id, bnull); //duplicate
						if(bnull || ngx_strncmp(params, "&threshold=", 11) != 0){
							sprintf(err_msg, "need logon before executing transactions");
							ret = -1;
						}else{
							params += 11;
							parse_param(params, sl_info->threshold, bnull);

							if(ngx_strncmp(params, "&trace=", 7) == 0){
								long long int ts = 0;
								bool tmp_bnull = false;
								params += 7;
								parse_param(params, ts, tmp_bnull);
								obc->trace_info.rte_send_ts = ts;
								assert(!tmp_bnull);
							}else{}

							if(bnull || *params != '\0' || sl_info->threshold < 1){
								sprintf(err_msg, "low level threshold invalid, should be positive integer");
								ret = -1;
							}else if( check_set_terminal_state(t_id, 1, log) ){ 
								ngx_log_error(NGX_LOG_DEBUG, log, 0, "stock_level:t_id=%d&w_id=%d&d_id=%d&threshold=%d",
												t_id, sl_info->w_id, sl_info->d_id, sl_info->threshold);
								if((ret = oceanbase_get_result(obc, log)) == OCEANBASE_RETURN_OK){
									ngx_log_error(NGX_LOG_WARN, log, 0, 
												"Immediate|stock_level:w_id=%d&d_id=%d&threshold=%d&low_stock=%d",
												t_id, sl_info->w_id, sl_info->d_id, sl_info->threshold, sl_info->low_stock);
									
								}else{}
							}else{
								obc->t_id = 0;
								ret = -2;
								sprintf(obc->msg, "%s%s stock_level transaction terminal %d is busy", 
																TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, t_id);
							}
						}
					}
				}
			}
		}
	}
	
	if(ret == -1){
		sprintf(obc->msg, "%s%s stock_level transaction parameters error: %s",
										TXN_EXEC_STATE_PREFIX, TXN_EXEC_FAILURE_MSG, err_msg);
	}else{}

	if(ret < 0){
		ret = OCEANBASE_RETURN_ERROR_CODE;
	}
	
	return ret;
}

ngx_int_t
ngx_http_oceanbase_output_stock_level(ngx_http_request_t *r, bool succ)
{
    u_char  *pos;
    char *out;
    ngx_http_upstream_t  *u = r->upstream;
    size_t  size;
    ngx_http_upstream_oceanbase_peer_data_t   *obp = u->peer.data;
    oceanbase_connection                          *obc = obp->oceanbase_con;
    int o_len = 0;
	ngx_log_t*  log = r->connection->log;
    StockLevelInfo * sl_info = &(obc->sl_info);
    size = STOCK_LEVEL_OUTPUT_BUF_SIZE;

    pos = ngx_http_oceanbase_request_mem(r, obp, size);
    if (pos == NULL) {
        return NGX_ERROR;
    }

    ngx_memzero(pos, size);

    out = (char *)pos;
    o_len += gen_response_head(out, "Stock Level");

	if(succ){
		o_len += sprintf(out + o_len, "Warehouse: %10d District: %2d <br>", sl_info->w_id, sl_info->d_id);
		o_len += sprintf(out + o_len, "Stock Level Threshold: %4d <br>", sl_info->threshold);
		o_len += sprintf(out + o_len, "Low Stock: %4d <br>", sl_info->low_stock);
	}else{}

    o_len += sprintf(out + o_len, "%s <br>", obc->msg);
    o_len += sprintf(out + o_len, txn_menu, sl_info->t_id, sl_info->d_id, sl_info->w_id,
                                            sl_info->t_id, sl_info->d_id, sl_info->w_id,
                                            sl_info->t_id, sl_info->d_id, sl_info->w_id,
                                            sl_info->t_id, sl_info->d_id, sl_info->w_id,
                                            sl_info->t_id, sl_info->d_id, sl_info->w_id);
    o_len += sprintf(out + o_len, "%s", response_footer);

    if(o_len >= (int)(size - 1)){
        ngx_log_error(NGX_LOG_ERR, log, 0, "stock-level output len: %d buffer size: %d\n", o_len, size);
        assert(false);
    }

    obp->seen_stream_end = 1;

    return ngx_http_oceanbase_submit_mem(r, obp, o_len + 1);
}

ngx_int_t
ngx_http_oceanbase_output_tpcc_txns(ngx_http_request_t *r, oceanbase_connection *obc, bool succ)
{
	ngx_int_t ret = 0;

	switch(obc->txn_type){
	case NEW_ORDER:
		ret = ngx_http_oceanbase_output_new_order(r, succ);
		break;
	case PAYMENT:
		ret = ngx_http_oceanbase_output_payment(r, succ);
		break;
	case DELIVERY:
		//ret = ngx_http_oceanbase_output_delivery(r, succ);
		assert(false);
		break;
	case ORDER_STATUS:
		ret = ngx_http_oceanbase_output_order_status(r, succ);
		break;
	case STOCK_LEVEL:
		ret = ngx_http_oceanbase_output_stock_level(r, succ);
		break;
	default:
		assert(false);
		break;
	}

	return ret;
}

int process_tpcc_txns(oceanbase_connection *obc, ngx_log_t *log )
{
	int ret = 0;
	
	ngx_log_error(NGX_LOG_DEBUG, log, 0, "txn_type:%d params:%s", obc->txn_type, obc->params);
	switch(obc->txn_type){
	case NEW_ORDER:
		ret = process_new_order(obc, log);
		break;
	case PAYMENT:
		ret = process_payment(obc, log);
		break;
	case DELIVERY:
		//ret = process_delivery(obc, log);
		assert(false);
		break;
	case ORDER_STATUS:
		ret = process_order_status(obc, log);
		break;
	case STOCK_LEVEL:
		ret = process_stock_level(obc, log);
		break;
	default:
		ret = -1;
		break;
	}

	return ret;
}

static int oceanbase_con_init_stmts(oceanbase_connection *obc)
{
    int ret = 0;
	char *msg = obc->msg;
    SQLRETURN retcode = SQL_SUCCESS;
    SQLHSTMT    no_stmt = NULL, pay_stmt = NULL, os_stmt = NULL, sl_stmt = NULL;

    memset(&(obc->no_param.ol_i_id_arr), 0, sizeof(obc->no_param.ol_i_id_arr));
    memset(&(obc->no_param.ol_s_w_id_arr), 0, sizeof(obc->no_param.ol_s_w_id_arr));
    memset(&(obc->no_param.i_price_arr), 0, sizeof(obc->no_param.i_price_arr));
    memset(&(obc->no_param.i_name_arr), 0, sizeof(obc->no_param.i_name_arr));
    memset(&(obc->no_param.s_qty_arr), 0, sizeof(obc->no_param.s_qty_arr));
    memset(&(obc->no_param.b_g_arr), 0, sizeof(obc->no_param.b_g_arr));
    memset(&(obc->no_param.ol_amount_arr), 0, sizeof(obc->no_param.ol_amount_arr));
    memset(&(obc->no_param.ol_qty_arr), 0, sizeof(obc->no_param.ol_qty_arr));
    memset(&(obc->no_param.sremote_arr), 0, sizeof(obc->no_param.sremote_arr));
    memset(&(obc->os_param.ol_i_id_arr_2), 0, sizeof(obc->os_param.ol_i_id_arr_2));
    memset(&(obc->os_param.ol_s_w_id_arr_2), 0, sizeof(obc->os_param.ol_s_w_id_arr_2));
    memset(&(obc->os_param.ol_amount_arr_2), 0, sizeof(obc->os_param.ol_amount_arr_2));
    memset(&(obc->os_param.ol_qty_arr_2), 0, sizeof(obc->os_param.ol_qty_arr_2));
    memset(&(obc->os_param.ol_del_d_arr_2), 0, sizeof(obc->os_param.ol_del_d_arr_2));

    if(SQL_FAIL(SQLAllocHandle(SQL_HANDLE_STMT, obc->dbc, &(obc->stmts[NEW_ORDER])))
         || !(no_stmt = obc->stmts[NEW_ORDER])){
        get_sql_diag(SQL_HANDLE_DBC, obc->dbc, msg, &obc->need_reconn);
        ret = -1;
    }else{
        int i = 0;
        for(i = 0; i < MAX_OL_NUM; i ++){
            obc->no_param.ol_i_id_num[i].type = SQL_C_LONG;
            obc->no_param.ol_i_id_num[i].is_null = false;
            obc->no_param.ol_i_id_num[i].data_buffer = &(obc->no_info.ol_i_id[i]);
            obc->no_param.ol_s_w_id_num[i].type = SQL_C_LONG;
            obc->no_param.ol_s_w_id_num[i].is_null = false;
            obc->no_param.ol_s_w_id_num[i].data_buffer = &(obc->no_info.ol_s_w_id[i]);
            obc->no_param.i_price_num[i].type = SQL_C_DOUBLE;
            obc->no_param.i_price_num[i].is_null = false;
            obc->no_param.i_price_num[i].data_buffer = &(obc->no_info.ol_i_price[i]);
            obc->no_param.i_name_str[i].type = SQL_C_CHAR;
            obc->no_param.i_name_str[i].is_null = false;
            obc->no_param.i_name_str[i].size = 25;
            obc->no_param.i_name_str[i].data_buffer = &(obc->no_info.ol_i_name[i]);
            obc->no_param.s_qty_num[i].type = SQL_C_LONG;
            obc->no_param.s_qty_num[i].is_null = false;
            obc->no_param.s_qty_num[i].data_buffer = &(obc->no_info.ol_stock[i]);
            obc->no_param.b_g_str[i].type = SQL_C_CHAR;
            obc->no_param.b_g_str[i].is_null = false;
            obc->no_param.b_g_str[i].size = 2;
            obc->no_param.b_g_str[i].data_buffer = &(obc->no_info.ol_b_g[i]);
            obc->no_param.ol_amount_num[i].type = SQL_C_DOUBLE;
            obc->no_param.ol_amount_num[i].is_null = false;
            obc->no_param.ol_amount_num[i].data_buffer = &(obc->no_info.ol_amount[i]);
            obc->no_param.ol_qty_num[i].type = SQL_C_LONG;
            obc->no_param.ol_qty_num[i].is_null = false;
            obc->no_param.ol_qty_num[i].data_buffer = &(obc->no_info.ol_quantity[i]);
            obc->no_param.sremote_num[i].type = SQL_C_LONG;
            obc->no_param.sremote_num[i].is_null = false;
            obc->no_param.sremote_num[i].data_buffer = &(obc->no_info.sremote[i]);
        }

        obc->no_param.ol_i_id_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.ol_i_id_arr.type_name = NULL;
        obc->no_param.ol_i_id_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.ol_i_id_arr.data_buffer = (void *)(obc->no_param.ol_i_id_num);
        obc->no_param.ol_s_w_id_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.ol_s_w_id_arr.type_name = NULL;
        obc->no_param.ol_s_w_id_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.ol_s_w_id_arr.data_buffer = (void *)(obc->no_param.ol_s_w_id_num);
        obc->no_param.i_price_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.i_price_arr.type_name = NULL;
        obc->no_param.i_price_arr.size =  MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.i_price_arr.data_buffer = (void *)(obc->no_param.i_price_num);
        obc->no_param.i_name_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.i_name_arr.type_name = NULL;
        obc->no_param.i_name_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.i_name_arr.data_buffer = (void *)(obc->no_param.i_name_str);
        obc->no_param.s_qty_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.s_qty_arr.type_name = NULL;
        obc->no_param.s_qty_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.s_qty_arr.data_buffer = (void *)(obc->no_param.s_qty_num);
        obc->no_param.b_g_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.b_g_arr.type_name = NULL;
        obc->no_param.b_g_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.b_g_arr.data_buffer = (void *)(obc->no_param.b_g_str);
        obc->no_param.ol_amount_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.ol_amount_arr.type_name = NULL;
        obc->no_param.ol_amount_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.ol_amount_arr.data_buffer = (void *)(obc->no_param.ol_amount_num);
        obc->no_param.ol_qty_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.ol_qty_arr.type_name = NULL;
        obc->no_param.ol_qty_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.ol_qty_arr.data_buffer = (void *)(obc->no_param.ol_qty_num);
        obc->no_param.sremote_arr.type = SQL_DRIVER_C_ARRAY;
        obc->no_param.sremote_arr.type_name = NULL;
        obc->no_param.sremote_arr.size = MAX_OL_NUM; //no_info.o_ol_cnt;
        obc->no_param.sremote_arr.data_buffer = (void *)(obc->no_param.sremote_num);


        if(SQL_FAIL(retcode = SQLPrepare(no_stmt, (SQLCHAR*)NEW_ORDER_PROC, SQL_NTS))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.w_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 2, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.d_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 3, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.c_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 4, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.alocal), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 5, SQL_PARAM_INPUT_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.o_ol_cnt), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 6, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                               4, 4, &(obc->no_info.w_tax), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 7, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                               4, 4, &(obc->no_info.d_tax), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 8, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.o_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 9, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                               4, 4, &(obc->no_info.c_discount), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 10, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               2, 0, &(obc->no_info.c_credit), 3, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 11, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               16, 0, &(obc->no_info.c_last), 17, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 12, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->no_info.retry), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 13, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR,
                               30, 0, obc->no_info.o_entry_d, 30, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 14, SQL_PARAM_INPUT, SQL_DRIVER_C_ARRAY,
                            SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.ol_i_id_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 15, SQL_PARAM_INPUT, SQL_DRIVER_C_ARRAY,
                            SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.ol_s_w_id_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 16, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                               SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.i_price_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 17, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                               SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.i_name_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 18, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                               SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.s_qty_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 19, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                               SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.b_g_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 20, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                    SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.ol_amount_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 21, SQL_PARAM_INPUT, SQL_DRIVER_C_ARRAY,
                               SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.sremote_arr), sizeof(ARRAY_STRUCT), NULL))
            || SQL_FAIL(retcode = SQLBindParameter(no_stmt, 22, SQL_PARAM_INPUT, SQL_DRIVER_C_ARRAY,
                               SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->no_param.ol_qty_arr), sizeof(ARRAY_STRUCT), NULL))){
            get_sql_diag(SQL_HANDLE_STMT, no_stmt, msg, &obc->need_reconn);
            ret = -1;
        }else if(SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_STMT, obc->dbc, &obc->stmts[PAYMENT]))
            || !(pay_stmt = obc->stmts[PAYMENT])){
            get_sql_diag(SQL_HANDLE_DBC, obc->dbc, msg, &obc->need_reconn);
            ret = -1;
        }else if(SQL_FAIL(retcode = SQLPrepare(pay_stmt, (SQLCHAR*)PAYMENT_PROC, SQL_NTS))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.w_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 2, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.d_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 3, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.c_w_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 4, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.c_d_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 5, SQL_PARAM_INPUT_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.c_id), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 6, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.blastname), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 7, SQL_PARAM_INPUT, SQL_C_DOUBLE, SQL_NUMERIC, 
                               6, 2, &(obc->pay_info.h_amount), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 8, SQL_PARAM_INPUT_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               16, 0, obc->pay_info.c_last, 17, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 9, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.w_street_1, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 10, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.w_street_2, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 11, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.w_city, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 12, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               2, 0, obc->pay_info.w_state, 3, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 13, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               9, 0, obc->pay_info.w_zip, 10, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 14, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.d_street_1, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 15, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.d_street_2, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 16, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.d_city, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 17, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               2, 0, obc->pay_info.d_state, 3, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 18, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               9, 0, obc->pay_info.d_zip, 10, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 19, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               16, 0, obc->pay_info.c_first, 17, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 20, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               2, 0, obc->pay_info.c_middle, 3, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 21, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.c_street_1, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 22, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.c_street_2, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 23, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               20, 0, obc->pay_info.c_city, 21, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 24, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               2, 0, obc->pay_info.c_state, 3, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 25, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               9, 0, obc->pay_info.c_zip, 10, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 26, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               16, 0, obc->pay_info.c_phone, 17, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 27, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               30, 0, obc->pay_info.c_since, 30, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 28, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               2, 0, obc->pay_info.c_credit, 3, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 29, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                               12, 2, &(obc->pay_info.c_credit_lim), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 30, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                               4, 4, &(obc->pay_info.c_discount), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 31, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                               12, 2, &(obc->pay_info.c_balance), 0, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 32, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                               200, 0, obc->pay_info.c_data, 201, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 33, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR,
                               30, 0, obc->pay_info.h_date, 30, NULL))
            || SQL_FAIL(retcode = SQLBindParameter(pay_stmt, 34, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                               0, 0, &(obc->pay_info.retry), 0, NULL))){
            get_sql_diag(SQL_HANDLE_STMT, pay_stmt, msg, &obc->need_reconn);
            ret = -1;
        }else{
            int i = 0;
            for(i = 0; i < MAX_OL_NUM; i ++){
                obc->os_param.ol_i_id_num_2[i].type = SQL_C_LONG;
                obc->os_param.ol_i_id_num_2[i].is_null = false;
                obc->os_param.ol_i_id_num_2[i].data_buffer = &(obc->os_info.ol_i_id[i]);
                obc->os_param.ol_s_w_id_num_2[i].type = SQL_C_LONG;
                obc->os_param.ol_s_w_id_num_2[i].is_null = false;
                obc->os_param.ol_s_w_id_num_2[i].data_buffer = &(obc->os_info.ol_s_w_id[i]);
                obc->os_param.ol_amount_num_2[i].type = SQL_C_DOUBLE;
                obc->os_param.ol_amount_num_2[i].is_null = false;
                obc->os_param.ol_amount_num_2[i].data_buffer = &(obc->os_info.ol_amount[i]);
                obc->os_param.ol_qty_num_2[i].type = SQL_C_LONG;
                obc->os_param.ol_qty_num_2[i].is_null = false;
                obc->os_param.ol_qty_num_2[i].data_buffer = &(obc->os_info.ol_quantity[i]);
                obc->os_param.ol_del_d_str_2[i].type = SQL_C_CHAR;
                obc->os_param.ol_del_d_str_2[i].is_null = false;
                obc->os_param.ol_del_d_str_2[i].size = 30;
                obc->os_param.ol_del_d_str_2[i].data_buffer = &(obc->os_info.ol_delivery_d[i]);
            }
            obc->os_param.ol_i_id_arr_2.type = SQL_DRIVER_C_ARRAY;
            obc->os_param.ol_i_id_arr_2.type_name = NULL;
            obc->os_param.ol_i_id_arr_2.size = MAX_OL_NUM;
            obc->os_param.ol_i_id_arr_2.data_buffer = (void *)(obc->os_param.ol_i_id_num_2);
            obc->os_param.ol_s_w_id_arr_2.type = SQL_DRIVER_C_ARRAY;
            obc->os_param.ol_s_w_id_arr_2.type_name = NULL;
            obc->os_param.ol_s_w_id_arr_2.size = MAX_OL_NUM;
            obc->os_param.ol_s_w_id_arr_2.data_buffer = (void *)(obc->os_param.ol_s_w_id_num_2);
            obc->os_param.ol_amount_arr_2.type = SQL_DRIVER_C_ARRAY;
            obc->os_param.ol_amount_arr_2.type_name = NULL;
            obc->os_param.ol_amount_arr_2.size = MAX_OL_NUM;
            obc->os_param.ol_amount_arr_2.data_buffer = (void *)(obc->os_param.ol_amount_num_2);
            obc->os_param.ol_qty_arr_2.type = SQL_DRIVER_C_ARRAY;
            obc->os_param.ol_qty_arr_2.type_name = NULL;
            obc->os_param.ol_qty_arr_2.size = MAX_OL_NUM;
            obc->os_param.ol_qty_arr_2.data_buffer = (void *)(obc->os_param.ol_qty_num_2);
            obc->os_param.ol_del_d_arr_2.type = SQL_DRIVER_C_ARRAY;
            obc->os_param.ol_del_d_arr_2.type_name = NULL;
            obc->os_param.ol_del_d_arr_2.size = MAX_OL_NUM;
            obc->os_param.ol_del_d_arr_2.data_buffer = (void *)(obc->os_param.ol_del_d_str_2);

            if(SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_STMT, obc->dbc, &(obc->stmts[ORDER_STATUS])))
                || !(os_stmt = obc->stmts[ORDER_STATUS])){
                get_sql_diag(SQL_HANDLE_DBC, obc->dbc, msg, &obc->need_reconn);
                ret = -1;
            }else if(SQL_FAIL(retcode = SQLPrepare(os_stmt, (SQLCHAR*)ORDER_STATUS_PROC, SQL_NTS))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.w_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 2, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.d_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 3, SQL_PARAM_INPUT_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.c_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 4, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.blastname), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 5, SQL_PARAM_INPUT_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                                   16, 0, obc->os_info.c_last, 17, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 6, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                                   16, 0, obc->os_info.c_first, 17, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 7, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                                   2, 0, obc->os_info.c_middle, 3, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 8, SQL_PARAM_OUTPUT, SQL_C_DOUBLE, SQL_NUMERIC,
                                   12, 2, &(obc->os_info.c_balance), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 9, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.o_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 10, SQL_PARAM_OUTPUT, SQL_C_CHAR, SQL_CHAR,
                                   30, 0, obc->os_info.o_entry_d, 30, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 11, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.o_carrier_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 12, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->os_info.o_ol_cnt), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 13, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                                 SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->os_param.ol_s_w_id_arr_2), sizeof(ARRAY_STRUCT), NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 14, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                                 SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->os_param.ol_i_id_arr_2), sizeof(ARRAY_STRUCT), NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 15, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                                   SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->os_param.ol_qty_arr_2), sizeof(ARRAY_STRUCT), NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 16, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                                 SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->os_param.ol_amount_arr_2), sizeof(ARRAY_STRUCT), NULL))
                || SQL_FAIL(retcode = SQLBindParameter(os_stmt, 17, SQL_PARAM_OUTPUT, SQL_DRIVER_C_ARRAY,
                                   SQL_DRIVER_SQL_TYPE_ARRAY, 0, 0, &(obc->os_param.ol_del_d_arr_2), sizeof(ARRAY_STRUCT), NULL))){
                get_sql_diag(SQL_HANDLE_STMT, os_stmt, msg, &obc->need_reconn);
                ret = -1;
            }else if(SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_STMT, obc->dbc, &(obc->stmts[STOCK_LEVEL])))
                || !(sl_stmt = obc->stmts[STOCK_LEVEL])){
                get_sql_diag(SQL_HANDLE_DBC, obc->dbc, msg, &obc->need_reconn);
                ret = -1;
            }else if(SQL_FAIL(retcode = SQLPrepare(sl_stmt, (SQLCHAR*)STOCK_LEVEL_PROC, SQL_NTS))
                || SQL_FAIL(retcode = SQLBindParameter(sl_stmt, 1, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->sl_info.w_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(sl_stmt, 2, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->sl_info.d_id), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(sl_stmt, 3, SQL_PARAM_INPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->sl_info.threshold), 0, NULL))
                || SQL_FAIL(retcode = SQLBindParameter(sl_stmt, 4, SQL_PARAM_OUTPUT, SQL_C_SLONG, SQL_INTEGER,
                                   0, 0, &(obc->sl_info.low_stock), 0, NULL))){
                get_sql_diag(SQL_HANDLE_STMT, sl_stmt, msg, &obc->need_reconn);
                ret = -1;
            }else{}
        }
    }

    if(ret < 0){
        if(no_stmt){
            SQLFreeHandle(SQL_HANDLE_STMT, no_stmt);
            obc->stmts[NEW_ORDER] = NULL;
        }
        if(pay_stmt){
            SQLFreeHandle(SQL_HANDLE_STMT, pay_stmt);
            obc->stmts[PAYMENT] = NULL;
        }
        if(os_stmt){
            SQLFreeHandle(SQL_HANDLE_STMT, os_stmt);
            obc->stmts[ORDER_STATUS] = NULL;
        }
        if(sl_stmt){
            SQLFreeHandle(SQL_HANDLE_STMT, sl_stmt);
            obc->stmts[STOCK_LEVEL] = NULL;
        }
    }
    return ret;
}

int total_conn_num = 0;

oceanbase_return_t ob_connect(oceanbase_connection *obc)
{
    int ret = 0;
    SQLRETURN retcode;
	char *msg = obc->msg;

    if(SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &(obc->env) ))){
        sprintf(msg, "SQLAllocHandle failed\n");
        ret = -1;
    }else if(SQL_FAIL(retcode = SQLSetEnvAttr(obc->env, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3_80, 0))
            || SQL_FAIL(retcode = SQLAllocHandle(SQL_HANDLE_DBC, obc->env, &(obc->dbc)))){
        get_sql_diag(SQL_HANDLE_ENV, obc->env, msg, &obc->need_reconn);
        ret = -1;
    }else if(SQL_FAIL(retcode = SQLConnect(obc->dbc, (SQLCHAR*)(obc->dsn), SQL_NTS, (SQLCHAR*)(obc->user),
                                        SQL_NTS, (SQLCHAR*)(obc->password), SQL_NTS))
            || SQL_FAIL(retcode = SQLSetConnectAttr(obc->dbc, SQL_ATTR_TXN_ISOLATION,
                                    (SQLPOINTER)SQL_TRANSACTION_SERIALIZABLE, SQL_IS_UINTEGER))){
        get_sql_diag(SQL_HANDLE_DBC, obc->dbc, msg, &obc->need_reconn);
        ret = -1;
    }else{
        ret = oceanbase_con_init_stmts(obc);
    }

    if(ret < 0){
        if(obc->dbc){
            SQLDisconnect(obc->dbc);
            SQLFreeHandle(SQL_HANDLE_DBC, obc->dbc);
            obc->dbc = NULL;
        }

        if(obc->env){
            SQLFreeHandle(SQL_HANDLE_ENV, obc->env);
            obc->env = NULL;
        }
    }else{
		obc->total_txn_cnt = 0;
        total_conn_num ++;
		if(!report_conn_num()){
        	sprintf(msg, "ob_connect report_conn_num failed\n");
		}else{}
    }

    return ret;
}

void ob_reset_conn(oceanbase_connection *obc)
{
    obc->txn_type = TXN_UNKNOWN;
    obc->estate = 0;
    obc->status = 0;
	memset(&(obc->trace_info), 0, sizeof(obc->trace_info));
}

int ob_free_conn(oceanbase_connection * obc)
{
	int ret = 0;
	assert(obc);

	if(obc->stmts[NEW_ORDER]){
		SQLFreeHandle(SQL_HANDLE_STMT, obc->stmts[NEW_ORDER]);
		obc->stmts[NEW_ORDER] = NULL;
	}
	if(obc->stmts[PAYMENT]){
		SQLFreeHandle(SQL_HANDLE_STMT, obc->stmts[PAYMENT]);
		obc->stmts[PAYMENT] = NULL;
	}
	if(obc->stmts[ORDER_STATUS]){
		SQLFreeHandle(SQL_HANDLE_STMT, obc->stmts[ORDER_STATUS]);
		obc->stmts[ORDER_STATUS] = NULL;
	}
	if(obc->stmts[STOCK_LEVEL]){
		SQLFreeHandle(SQL_HANDLE_STMT, obc->stmts[STOCK_LEVEL]);
		obc->stmts[STOCK_LEVEL] = NULL;
	}

	if(obc->dbc){
		SQLDisconnect(obc->dbc);
		SQLFreeHandle(SQL_HANDLE_DBC, obc->dbc);
		obc->dbc = NULL;
	}

	if(obc->env){
		SQLFreeHandle(SQL_HANDLE_ENV, obc->env);
		obc->env = NULL;
	}

	total_conn_num --;
	if(!report_conn_num()){
        sprintf(obc->msg, "ob_free_conn report_conn_num failed\n");
	}else{}
    
	return ret;
}

int ob_get_conn_fd(oceanbase_connection * obc)
{
    SQLRETURN retcode;
    int fd = -1;

    if(SQL_FAIL(retcode = SQLGetAsyncConnFd(obc->dbc, &fd))){
        printf("SQLGetAsyncConnFd failed\n");
    }else{}

    return fd;
}
