/*
 * ngx_tpcc_utils.c contains support routines for test
 *
 * Copyright (C) 2019 Ant Financial (Hang Zhou) Network Technology Co., Ltd.
 *
 */
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>
#include "ngx_tpcc_common.h"

void get_sql_diag( SQLSMALLINT type, SQLHANDLE handle, char *out_state, bool * reconn)
{
    SQLCHAR     state[6];
    SQLINTEGER  errno;
    SQLSMALLINT len = 0, s_len = 0;
    SQLCHAR     message[SQL_MAX_MESSAGE_LENGTH];
    SQLSMALLINT idx = 1;
    SQLRETURN   ret;

    while(1)
    {
        ret = SQLGetDiagRec(type, handle, idx, state, &errno, message, sizeof(message), &len);

        if( ret == SQL_NO_DATA ){
            break;
        }else if (ret == SQL_SUCCESS){
			if((strcmp((char *)state, "08S01") == 0) && (reconn != NULL)){
				*reconn = true;
			}else{}

			s_len += sprintf(out_state + s_len, "SQLSTATE:%s NATIVE:%d MSG:%s", state, errno, message);
            idx ++;
        }else{
            s_len += sprintf(out_state + s_len, "SQLGetDiagRec failed, SQLRETURN: %d\n", ret);
            break;
        }
    }
}

bool get_timestamp(char * str)
{
    bool ret = true;
    struct tm *lt = NULL, lt_;
    struct timeval tv;

    if(gettimeofday(&tv, NULL) == 0){
		tzset();
        lt = localtime_r(&tv.tv_sec, &lt_);
        if(lt != NULL){
            sprintf(str, "%04d-%02d-%02d %02d:%02d:%02d",
                   lt->tm_year + 1900,
                   lt->tm_mon + 1,
                   lt->tm_mday,
                   lt->tm_hour,
                   lt->tm_min,
                   lt->tm_sec);
        }else{
            ret = false;
        }
    }else{
        ret = false;
    }

    return ret;
}

bool SQL_SUCC(int retcode)
{
    return (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO || retcode == SQL_STILL_EXECUTING);
}

bool SQL_FAIL(int retcode)
{
 	return !SQL_SUCC(retcode);
}

bool rollback_on_error(SQLHANDLE dbc, char* msg, bool* need_reconn)
{
	bool ret = true;

    if(SQL_FAIL(SQLEndTran(SQL_HANDLE_DBC, dbc, SQL_ROLLBACK))){
        get_sql_diag(SQL_HANDLE_DBC, dbc, msg, need_reconn);
		ret = false;
    }else{}
	
	return ret;
}
